
from vsg.rules import previous_line

from vsg import token

lTokens = []
lTokens.append(token.library_clause.keyword)


class rule_003(previous_line):
    '''
    This rule checks for blank lines or comments above the **library** keyword.

    |configuring_previous_line_rules_link|

    There is an additional **allow_library_clause** option which can be set.
    Refer to section :ref:`reporting-single-rule-configuration` for details on finding configuration options for individual rules.

    allow_library_clause
    ^^^^^^^^^^^^^^^^^^^^

    When set to **True**, it allows consecutive library clauses.

    **Violation**

    .. code-block:: vhdl

       library ieee;
         use ieee.std_logic_1164.all;
       library top_dsn;
       library fifo_dsn;

    **Fix**

    .. code-block:: vhdl

       library ieee;
         use ieee.std_logic_1164.all;

       library top_dsn;
       library fifo_dsn;
    '''

    def __init__(self):
        previous_line.__init__(self, 'library', '003', lTokens)
        self.allow_library_clause = False
        self.configuration.append('allow_library_clause')
        self.style = 'require_blank_line'

    def _set_allow_tokens(self):
        if self.allow_library_clause:
            self.lAllowTokens = []
            self.lAllowTokens.append(token.library_clause.keyword)




from vsg.token import logical_name_list as token

from vsg.vhdlFile import utils


def classify_until(lUntils, iToken, lObjects):
    '''
    logical_name_list ::=
        logical_name { , logical_name }
    '''
    iCurrent = iToken
    iLast = 0
    while iLast != iCurrent:
        iLast = iCurrent
        if lObjects[utils.find_next_token(iCurrent, lObjects)].get_value().lower() in lUntils:
            return iCurrent
        iCurrent = utils.assign_next_token_if(',', token.comma, iCurrent, lObjects)
        iCurrent = utils.assign_next_token(token.logical_name, iCurrent, lObjects)




from vsg import parser


class keyword(parser.keyword):
    '''
    unique_id = type_declaration : keyword
    '''

    def __init__(self, sString):
        parser.keyword.__init__(self, sString)


class is_keyword(parser.keyword):
    '''
    unique_id = type_declaration : is_keyword
    '''

    def __init__(self, sString):
        parser.keyword.__init__(self, sString)


class type_definition(parser.item):
    '''
    unique_id = type_declaration : type_definition
    '''

    def __init__(self, sString):
        parser.item.__init__(self, sString)


class semicolon(parser.semicolon):
    '''
    unique_id = type_declaration : semicolon
    '''

    def __init__(self):
        parser.semicolon.__init__(self)




from vsg.rules import blank_line_below_line_ending_with_token as Rule

from vsg.token import entity_declaration as token


class rule_203(Rule):
    '''
    This rule checks for blank lines below the semicolon in entity specifications.

    |configuring_blank_lines_link|

    **Violation**

    .. code-block:: vhdl

       end entity;
       library ieee;

    **Fix**

    .. code-block:: vhdl

       end entity;

       library ieee;
    '''

    def __init__(self):
        Rule.__init__(self, 'entity', '203', [token.semicolon])




from vsg.token import group_declaration as token

from vsg.vhdlFile import utils

from vsg.vhdlFile.classify import group_constituent_list


def detect(iToken, lObjects):
    '''
    group_declaration ::=
        group identifier : group_template_name ( group_constituent_list ) ;
    '''

    if utils.is_next_token('group', iToken, lObjects):
        return classify(iToken, lObjects)
    return iToken


def classify(iToken, lObjects):

    iCurrent = utils.assign_next_token_required('group', token.group_keyword, iToken, lObjects)
    iCurrent = utils.assign_next_token(token.identifier, iCurrent, lObjects)
    iCurrent = utils.assign_next_token_required(':', token.colon, iCurrent, lObjects)
    iCurrent = utils.assign_next_token(token.group_template_name, iCurrent, lObjects)
    iCurrent = utils.assign_next_token_required('(', token.open_parenthesis, iCurrent, lObjects)
    iCurrent = group_constituent_list.classify(iCurrent, lObjects)
    iCurrent = utils.assign_next_token_required(')', token.close_parenthesis, iCurrent, lObjects)
    iCurrent = utils.assign_next_token_required(';', token.semicolon, iCurrent, lObjects)

    return iCurrent




from vsg.token import record_type_definition as token

from vsg.vhdlFile import utils

from vsg.vhdlFile.classify import element_declaration


def detect(iToken, lObjects):
    '''
    record_type_definition ::=
        record
            element_declaration
            { element_declaration }
        end record [ *record_type*_simple_name ]
    '''

    if utils.is_next_token('record', iToken, lObjects):
        return classify(iToken, lObjects)

    return iToken


def classify(iToken, lObjects):

    iCurrent = utils.assign_next_token_required('record', token.record_keyword, iToken, lObjects)

    iCurrent = utils.classify_subelement_until('end', element_declaration, iCurrent, lObjects)

    iCurrent = utils.assign_next_token_required('end', token.end_keyword, iCurrent, lObjects)
    iCurrent = utils.assign_next_token_required('record', token.end_record_keyword, iCurrent, lObjects)
    iCurrent = utils.assign_next_token_if_not(';', token.record_type_simple_name, iCurrent, lObjects)

    return iCurrent




from vsg.rules import token_indent

from vsg import token

lTokens = []
lTokens.append(token.package_body.end_keyword)


class rule_301(token_indent):
    '''
    This rule checks the indent of the end package declaration.

    **Violation**

    .. code-block:: vhdl

       package body FIFO_PKG is

          end package body fifo_pkg;

    **Fix**

    .. code-block:: vhdl

       package body fifo_pkg is

       end package body fifo_pkg;
    '''

    def __init__(self):
        token_indent.__init__(self, 'package_body', '301', lTokens)




from vsg.rules import token_indent

from vsg import token

lTokens = []
lTokens.append(token.component_declaration.component_keyword)


class rule_001(token_indent):
    '''
    This rule checks the indentation of the **component** keyword.

    **Violation**

    .. code-block:: vhdl

       architecture rtl of fifo is
       begin

       component fifo is

            component ram is

    **Fix**

    .. code-block:: vhdl

       architecture rtl of fifo is
       begin

         component fifo is

         component ram is
    '''

    def __init__(self):
        token_indent.__init__(self, 'component', '001', lTokens)




from vsg.vhdlFile.extract import tokens


def get_tokens_matching(lTokens, lAllTokens, oTokenMap):

    lReturn = []
    lIndexes = []

    for oToken in lTokens:
        lIndexes.extend(oTokenMap.get_token_indexes(oToken))

    lIndexes.sort()

    for iIndex in lIndexes:
        iLine = oTokenMap.get_line_number_of_index(iIndex)
        lReturn.append(tokens.New(iIndex, iLine, [lAllTokens[iIndex]]))

    return lReturn




from vsg.rules import token_indent_between_tokens

from vsg import token

lTokens = []
lTokens.append(token.interface_constant_declaration.constant_keyword)
lTokens.append(token.interface_signal_declaration.signal_keyword)
lTokens.append(token.interface_variable_declaration.variable_keyword)
lTokens.append(token.interface_file_declaration.file_keyword)
lTokens.append(token.interface_unknown_declaration.identifier)
lTokens.append(token.interface_incomplete_type_declaration.type_keyword)
lTokens.append(token.interface_procedure_specification.procedure_keyword)
lTokens.append(token.interface_function_specification.function_keyword)
lTokens.append(token.interface_package_declaration.package_keyword)

oStart = token.procedure_specification.open_parenthesis
oEnd = token.procedure_specification.close_parenthesis


class rule_004(token_indent_between_tokens):
    '''
    This rule checks the indent of parameters.

    **Violation**

    .. code-block:: vhdl

       procedure average_samples (
       constant a : in integer;
           signal b : in std_logic;
          variable c : in std_logic_vector(3 downto 0);
        signal d : out std_logic ) is
       begin
       end procedure average_samples;

    **Fix**

    .. code-block:: vhdl

       procedure average_samples (
         constant a : in integer;
         signal b : in std_logic;
         variable c : in std_logic_vector(3 downto 0);
         signal d : out std_logic ) is
       begin
       end procedure average_samples;
    '''

    def __init__(self):
        token_indent_between_tokens.__init__(self, 'procedure', '004', lTokens, oStart, oEnd)




from vsg import parser

from vsg.rules import remove_spaces_before_token_rule


class rule_006(remove_spaces_before_token_rule):
    '''
    This rule checks for spaces before a close parenthesis.

    **Violation**

    .. code-block:: vhdl

       signal data        : std_logic_vector(31 downto 0    );
       signal byte_enable : std_logic_vector( 3 downto 0 );
       signal width       : std_logic_vector(g_width - 1 downto 0);

    **Fix**

    .. code-block:: vhdl

       signal data        : std_logic_vector(31 downto 0);
       signal byte_enable : std_logic_vector( 3 downto 0);
       signal width       : std_logic_vector(g_width - 1 downto 0);
    '''

    def __init__(self):
        remove_spaces_before_token_rule.__init__(self, 'whitespace', '006', parser.close_parenthesis, bIgnoreIfLineStart=True)
        self.solution = 'Remove spaces before close ).'
        self.configuration_documentation_link = None




from vsg.token import subtype_declaration as token

from vsg.vhdlFile import utils

from vsg.vhdlFile.classify import identifier
from vsg.vhdlFile.classify import subtype_indication



def detect(iToken, lObjects):
    '''
    subtype_declaration ::=
        subtype identifier is subtype_indication ;
    '''

    if utils.is_next_token('subtype', iToken, lObjects):
        return classify(iToken, lObjects)

    return iToken


def classify(iToken, lObjects):

    iCurrent = utils.assign_next_token_required('subtype', token.subtype_keyword, iToken, lObjects)

    iCurrent = identifier.classify(iCurrent, lObjects, token.identifier)

    iCurrent = utils.assign_next_token_required('is', token.is_keyword, iCurrent,lObjects)

    iCurrent = subtype_indication.classify(iCurrent, lObjects)

    iCurrent = utils.assign_next_token_required(';', token.semicolon, iCurrent, lObjects)

    return iCurrent




from vsg.rules import insert_token_right_of_token_if_it_does_not_exist_between_tokens_using_value_from_token as Rule

from vsg import token

oInsertToken = token.subprogram_kind.procedure_keyword

oAnchorToken = token.subprogram_body.end_keyword

oLeftToken = token.procedure_specification.procedure_keyword
oRightToken = token.subprogram_body.semicolon

oValueToken = token.procedure_specification.procedure_keyword


class rule_012(Rule):
    '''
    This rule checks the procedure keyword exist in the closing of the procedure specification.

    |configuring_optional_items_link|

    **Violation**

    .. code-block:: vhdl

       procedure proc is

       end proc;

    **Fix**

    .. code-block:: vhdl

       procedure proc is

       end procedure proc;
    '''

    def __init__(self):
        Rule.__init__(self, 'procedure', '012', oInsertToken, oAnchorToken, oLeftToken, oRightToken, oValueToken)
        self.solution = 'procedure keyword'
        self.groups.append('structure::optional')
        self.filter_tokens.append(token.subprogram_declaration.semicolon)




from vsg.token import selected_variable_assignment as token

from vsg.vhdlFile import utils

from vsg.vhdlFile.classify import expression
from vsg.vhdlFile.classify import selected_expressions


def detect(iToken, lObjects):
    '''
    selected_variable_assignment ::=
        with expression select [ ? ]
           target := selected_expressions ;
    '''

    if utils.is_next_token_one_of(['when', 'if', 'elsif', 'else'], iToken, lObjects):
        return False
    if utils.find_in_range(':=', iToken, ';', lObjects):
        if utils.find_in_range('with', iToken, ';', lObjects):
            return True
        return False
    return False


def classify(iToken, lObjects):

    iCurrent = utils.assign_next_token_required('with', token.with_keyword, iToken, lObjects)

    iCurrent = expression.classify_until(['select'], iToken, lObjects)

    iCurrent = utils.assign_next_token_required('select', token.select_keyword, iToken, lObjects)
    iCurrent = utils.assign_next_token_if('?', token.question_mark, iCurrent, lObjects)
    iCurrent = utils.assign_tokens_until(':=', token.target, iCurrent, lObjects)
    iCurrent = utils.assign_next_token_required(':=', token.assignment, iCurrent, lObjects)

    iCurrent = selected_expressions.classify_until([';'], iToken, lObjects)

    iCurrent = utils.assign_next_token_required(';', token.semicolon, iCurrent, lObjects)

    return iCurrent




from vsg import token

from vsg.rules.whitespace_after_token import Rule

lTokens = []
lTokens.append(token.variable_declaration.colon)


class rule_005(Rule):
    '''
    This rule checks there is a single space after the colon.

    |configuring_whitespace_rules_link|

    **Violation**

    .. code-block:: vhdl

       variable count   :integer;
       variable counter :     integer;

    **Fix**

    .. code-block:: vhdl

       variable count   : integer;
       variable counter : integer;
    '''
    def __init__(self):
        Rule.__init__(self, 'variable', '005', lTokens)




from vsg.rules import insert_carriage_return_after_token_if_it_is_not_followed_by_a_comment

from vsg import token

lTokens = []
lTokens.append(token.simple_waveform_assignment.semicolon)
lTokens.append(token.simple_force_assignment.semicolon)
lTokens.append(token.simple_release_assignment.semicolon)


class rule_007(insert_carriage_return_after_token_if_it_is_not_followed_by_a_comment):
    '''
    This rule checks for code after a sequential assignment.

    **Violation**

    .. code-block:: vhdl

        a <= '0'; b <= '1'; c <= '0'; -- comment

    **Fix**

    .. code-block:: vhdl

        a <= '0';
        b <= '1';
        c <= '0'; -- comment
    '''

    def __init__(self):
        insert_carriage_return_after_token_if_it_is_not_followed_by_a_comment.__init__(self, 'sequential', '007', lTokens)
        self.solution = 'Move code after the ; to the next line.'




from vsg import token

from vsg.rules import formal_part_in_association_element_between_tokens

lStart = token.port_map_aspect.open_parenthesis
lEnd = token.port_map_aspect.close_parenthesis


class rule_008(formal_part_in_association_element_between_tokens):
    '''
    This rule checks for positional ports.
    Positional ports are subject to problems when the position of the underlying component changes.

    **Violation**

    .. code-block:: vhdl

       port map (
         WR_EN, RD_EN, OVERFLOW
       );

    **Fix**

    Use explicit port mapping.

    .. code-block:: vhdl

       port map (
         WR_EN    => WR_EN,
         RD_EN    => RD_EN,
         OVERFLOW => OVERFLOW
       );
    '''
    def __init__(self):
        formal_part_in_association_element_between_tokens.__init__(self, 'port_map', '008', lStart, lEnd)
        self.solution = 'Add formal_part to positional assignment.'




from vsg.token import association_list as token

from vsg.vhdlFile import utils

from vsg.vhdlFile.classify import association_element


def classify(iToken, lObjects):
    '''
    association_list ::=
        association_element { , association_element }
    '''
    iCurrent = association_element.detect(iToken, lObjects)

    while utils.is_next_token(',', iCurrent, lObjects):
        iCurrent = utils.assign_next_token_required(',', token.comma, iCurrent, lObjects)
        iCurrent = association_element.detect(iCurrent, lObjects)

    return iCurrent




from vsg.vhdlFile.classify import interface_list


def classify(iToken, lObjects):
    '''
    generic_list ::=
        *generic*_interface_list
    '''

    return interface_list.classify(iToken, lObjects)




from vsg.rules import split_line_at_token as Rule

from vsg import token

lTokens = []
lTokens.append(token.alias_declaration.alias_keyword)


class rule_001(Rule):
    '''
    This rule checks the **alias** keyword is on its own line.

    **Violation**

    .. code-block:: vhdl

       signal sig1 : std_logic; alias designator is name;

    **Fix**

    .. code-block:: vhdl

       signal sig1 : std_logic;
       alias designator is name;
    '''

    def __init__(self):
        Rule.__init__(self, 'alias_declaration', '001', lTokens)
        self.solution = 'Move *alias* keyword and code after end to the next line'




from vsg.rules import move_token as Rule

from vsg import token

oToken = token.record_type_definition.record_keyword


class rule_001(Rule):
    '''
    This rule checks the location of the **record** keyword.

    The default location is not on a line by itself.

    |configuring_move_token_rules_link|

    **Violation**

    .. code-block:: vhdl

       type t_record is
       record

    **Fix**

    .. code-block:: vhdl

       type t_record is record
    '''

    def __init__(self):
        Rule.__init__(self, 'record_type_definition', '001', oToken)
        self.action = 'same_line'
        self.subphase = 3
        self.insert_whitespace = True




from vsg.rules import insert_carriage_return_after_token_if_it_is_not_followed_by_a_comment_when_between_tokens as Rule

from vsg import token

lTokens = []
lTokens.append(token.delay_mechanism.transport_keyword)
lTokens.append(token.delay_mechanism.inertial_keyword)

lTokenPairs = []
lTokenPairs.append([token.concurrent_selected_signal_assignment.with_keyword, token.concurrent_selected_signal_assignment.semicolon])
lTokenPairs.append([token.selected_waveform_assignment.with_keyword, token.selected_waveform_assignment.semicolon])


class rule_009(Rule):
    '''
    This rule checks for code after the delay mechanism keywords **transport** and **inertial**.

    **Violation**

    .. code-block:: vhdl

       with mux_sel select addr <= transport "0000" when 0,
         "0001" when 1,
         "1111" when others;

    **Fix**

    .. code-block:: vhdl

       with mux_sel select addr <= transport
         "0000" when 0,
         "0001" when 1,
         "1111" when others;
    '''

    def __init__(self):
        Rule.__init__(self, 'selected_assignment', '009', lTokens, lTokenPairs)
        self.solution = 'Move code after the delay mechanism keyword to the next line.'




from vsg.rules import blank_line_below_line_ending_with_token

from vsg.token import case_statement as token


class rule_010(blank_line_below_line_ending_with_token):
    '''
    This rule checks for a blank line below the **end case** keywords.

    |configuring_blank_lines_link|

    **Violation**

    .. code-block:: vhdl

       end case;
       a <= '1';

    **Fix**

    .. code-block:: vhdl

       end case;

       a <= '1';
    '''

    def __init__(self):
        blank_line_below_line_ending_with_token.__init__(self, 'case', '010', [token.semicolon])




from vsg.rules import split_line_at_token

from vsg import token

lTokens = []
lTokens.append(token.context_reference.keyword)


class rule_005(split_line_at_token):
    '''
    This rule checks the **context** keyword is on its own line.

    **Violation**

    .. code-block:: vhdl

       context c1 is library ieee; context con1; end context c1;

       library ieee; context con2;

    **Fix**

    .. code-block:: vhdl

       context c1 is library ieee;
       context con1; end context c1;

       library ieee;
       context con2;
    '''

    def __init__(self):
        split_line_at_token.__init__(self, 'context_ref', '005', lTokens)
        self.solution = 'Move context and code after context to the next line'




from vsg import token

from vsg.rules import token_suffix

lTokens = []
lTokens.append(token.package_declaration.identifier)
lTokens.append(token.package_declaration.end_package_simple_name)


class rule_016(token_suffix):
    '''
    This rule checks for valid suffixes on package identifiers.
    The default package suffix is *_pkg*.

    |configuring_prefix_and_suffix_rules_link|

    **Violation**

    .. code-block:: vhdl

       package foo is

    **Fix**

    .. code-block:: vhdl

       package foo_pkg is
    '''

    def __init__(self):
        token_suffix.__init__(self, 'package', '016', lTokens)
        self.suffixes = ['_pkg']




from vsg.rules import token_case

from vsg import token

lTokens = []
lTokens.append(token.instantiated_unit.component_keyword)


class rule_031(token_case):
    '''
    This rule checks the component keyword has proper case in component instantiations that use the **component** keyword.

    |configuring_uppercase_and_lowercase_rules_link|

    **Violation**

    .. code-block:: vhdl

       instance_name : COMPONENT entity_name

    **Fix**

    .. code-block:: vhdl

       instance_name : component entity_name
    '''

    def __init__(self):
        token_case.__init__(self, 'instantiation', '031', lTokens)
        self.groups.append('case::keyword')




from vsg.rules import move_token as Rule

from vsg import token

oToken = token.generic_map_aspect.close_parenthesis


class rule_004(Rule):
    '''
    This rule checks the location of the closing ")" character for the generic map.

    The default location is on a line by itself.

    |configuring_move_token_rules_link|

    **Violation**

    .. code-block:: vhdl

         generic map (
           GENERIC_1 => 0,
           GENERIC_2 => TRUE,
           GENERIC_3 => FALSE)

    **Fix**

    .. code-block:: vhdl

         generic map (
           GENERIC_1 => 0,
           GENERIC_2 => TRUE,
           GENERIC_3 => FALSE
         )
    '''

    def __init__(self):
        Rule.__init__(self, 'generic_map', '004', oToken)





from vsg.rules import move_token_right_to_next_non_whitespace_token as Rule

from vsg import token

lTokens = []
lTokens.append(token.process_statement.postponed_keyword)


class rule_039(Rule):
    '''
    This rule checks a **postponed** keyword is on the same line at the **process** keyword.

    **Violation**

    .. code-block:: vhdl

       label : postponed
       process

    **Fix**

    .. code-block:: vhdl

       label :
       postponed process
    '''

    def __init__(self):
        Rule.__init__(self, 'process', '039', lTokens)




from vsg.rules import token_indent

from vsg import token

lTokens = []
lTokens.append(token.context_declaration.context_keyword)


class rule_001(token_indent):
    '''
    This rule checks the indent of the **context** keyword.

    **Violation**

    .. code-block:: vhdl

         context c1 is

         library ieee;

    **Fix**

    .. code-block:: vhdl

       context c1 is

         library ieee;
    '''

    def __init__(self):
        token_indent.__init__(self, 'context', '001', lTokens)




from vsg.vhdlFile.classify import procedure_specification
from vsg.vhdlFile.classify import function_specification


def detect(iCurrent, lObjects):
    '''
    subprogram_specification ::=
        procedure_specification
      | function_specification
    '''

    iReturn = procedure_specification.detect(iCurrent, lObjects)
    if iReturn != iCurrent:
        return iReturn

    iReturn = function_specification.detect(iCurrent, lObjects)
    if iReturn != iCurrent:
        return iReturn

    return iCurrent




from vsg.rules import blank_line_below_line_ending_with_token

from vsg import token

lTokens = []
lTokens.append(token.if_statement.semicolon)


class rule_030(blank_line_below_line_ending_with_token):
    '''
    This rule checks a single blank line after the **end if**.
    In the case of nested **if** statements, the rule will be enfoced on the last **end if**.

    |configuring_blank_lines_link|

    **Violation**

    .. code-block:: vhdl

       if (A = '1') then
         B <= '0';
       end if;
       C <= '1';

    **Fix**

    .. code-block:: vhdl

       if (A = '1') then
         B <= '0';
       end if;

       C <= '1';
    '''

    def __init__(self):
        blank_line_below_line_ending_with_token.__init__(self, 'if', '030', lTokens)
        self.lHierarchyLimits = [0]




from vsg.rules import blank_line_below_line_ending_with_token

from vsg import token

lTokens = []
lTokens.append(token.block_statement.semicolon)


class rule_205(blank_line_below_line_ending_with_token):
    '''
    This rule checks for a blank line below the semicolon.

    |configuring_blank_lines_link|

    **Violation**

    .. code-block:: vhdl

       end block block_label;
       a <= b;

    **Fix**

    .. code-block:: vhdl

       end block block_label;

       a <= b;
    '''

    def __init__(self):
        blank_line_below_line_ending_with_token.__init__(self, 'block', '205', lTokens)




from vsg import parser


class std_logic_vector(parser.type):
    '''
    unique_id = ieee_std_logic_1164_types : std_logic_vector
    '''

    def __init__(self, sString):
        parser.type.__init__(self, sString)


class std_ulogic_vector(parser.type):
    '''
    unique_id = ieee_std_logic_1164_types : std_ulogic_vector
    '''

    def __init__(self, sString):
        parser.type.__init__(self, sString)


class std_ulogic(parser.type):
    '''
    unique_id = ieee_std_logic_1164_types : std_ulogic
    '''

    def __init__(self, sString):
        parser.type.__init__(self, sString)


class std_logic(parser.type):
    '''
    unique_id = ieee_std_logic_1164_types : std_logic
    '''

    def __init__(self, sString):
        parser.type.__init__(self, sString)


class integer(parser.type):
    '''
    unique_id = ieee_std_logic_1164_types : integer
    '''

    def __init__(self, sString):
        parser.type.__init__(self, sString)


class signed(parser.type):
    '''
    unique_id = ieee_std_logic_1164_types : signed
    '''

    def __init__(self, sString):
        parser.type.__init__(self, sString)


class unsigned(parser.type):
    '''
    unique_id = ieee_std_logic_1164_types : unsigned
    '''

    def __init__(self, sString):
        parser.type.__init__(self, sString)


class natural(parser.type):
    '''
    unique_id = ieee_std_logic_1164_types : natural
    '''

    def __init__(self, sString):
        parser.type.__init__(self, sString)




from vsg import parser
from vsg import violation

from vsg.rule_group import structure
from vsg.rules import utils as rules_utils


class insert_token_left_of_token_if_it_does_not_exist_between_tokens(structure.Rule):
    '''
    Checks for the existence of a token and will insert it if it does not exist.

    Parameters
    ----------

    name : string
       The group the rule belongs to.

    identifier : string
       unique identifier.  Usually in the form of 00N.

    insert_token : token object
       token to insert if it does not exist.

    anchor_token : token object type
       token to check if insert_token exists to the right of

    end_token : token object type
       token that bounds the search for the insert_token
    '''

    def __init__(self, name, identifier, insert_token, anchor_token, start_token, end_token):
        structure.Rule.__init__(self, name=name, identifier=identifier)
        self.insert_token = insert_token
        self.anchor_token = anchor_token
        self.start_token = start_token
        self.end_token = end_token
        self.action = 'add'
        self.configuration.append('action')
        self.configuration_documentation_link = 'configuring_optional_items_link'

    def _get_tokens_of_interest(self, oFile):
        if self.action == 'remove':
            return oFile.get_token_and_n_tokens_before_it([self.insert_token], 1)
        else:
            return oFile.get_tokens_bounded_by(self.start_token, self.end_token)

    def _analyze(self, lToi):
        if self.action == 'remove':
            for oToi in lToi:
                sSolution = self.action.capitalize() + ' ' + self.solution
                self.add_violation(violation.New(oToi.get_line_number(), oToi, sSolution))
            return

        for oToi in lToi:
            lTokens = oToi.get_tokens()
            dAction = {}
            for iToken, oToken in enumerate(lTokens):
                if isinstance(oToken, type(self.insert_token)):
                    break
                if isinstance(oToken, self.anchor_token):
                    dAction['index'] = iToken
            else:
                if dAction == {}:
                    continue
                sSolution = self.action.capitalize() + ' ' + self.solution
                oViolation  = violation.New(oToi.get_line_number(), oToi, sSolution)
                oViolation.set_action(dAction)
                self.add_violation(oViolation)

    def _fix_violation(self, oViolation):
        if self.action == 'remove':
            rules_utils.remove_optional_item(oViolation, self.insert_token)
        else:
            lTokens = oViolation.get_tokens()
            dAction = oViolation.get_action()
            rules_utils.insert_token(lTokens, dAction['index'], self.insert_token)
            rules_utils.insert_whitespace(lTokens, dAction['index'] + 1)
            oViolation.set_tokens(lTokens)




from vsg import token

from vsg.rules.whitespace_after_token import Rule

lTokens = []
lTokens.append(token.alias_declaration.is_keyword)


class rule_102(Rule):
    '''
    This rule checks for a single space after the **is** keyword.

    |configuring_whitespace_rules_link|

    **Violation**

    .. code-block:: vhdl

       alias alias_designator is     name;

    **Fix**

    .. code-block:: vhdl

       alias alias_designator is name;
    '''
    def __init__(self):
        Rule.__init__(self, 'alias_declaration', '102', lTokens)




from vsg.rules import insert_carriage_return_after_token_if_it_is_not_followed_by_a_comment_when_between_tokens as Rule

from vsg import token

lTokens = []
lTokens.append(token.interface_list.semicolon)

lTokenPairs = []
lTokenPairs.append([token.port_clause.open_parenthesis, token.port_clause.close_parenthesis])


class rule_013(Rule):
    '''
    This rule checks for multiple ports declared on a single line.

    **Violation**

    .. code-block:: vhdl

       port (
         WR_EN    : in    std_logic;RD_EN    : in    std_logic;
         OVERFLOW : out   std_logic;DATA     : inout std_logic
       );

    **Fix**

    .. code-block:: vhdl

       port (
         WR_EN    : in    std_logic;
         RD_EN    : in    std_logic;
         OVERFLOW : out   std_logic;
         DATA     : inout std_logic
       );
    '''

    def __init__(self):
        Rule.__init__(self, 'port', '013', lTokens, lTokenPairs)
        self.solution = 'Move multiple ports to their own lines.'





from vsg import parser
from vsg import violation

from vsg.rule_group import structure
from vsg.rules import utils as rules_utils
from vsg.vhdlFile import utils


class move_token(structure.Rule):
    '''
    Checks the case for words.

    Parameters
    ----------

    name : string
       The group the rule belongs to.

    identifier : string
       unique identifier.  Usually in the form of 00N.

    oToken : token type
       object type to split a line at
    '''

    def __init__(self, name, identifier, oToken):
        structure.Rule.__init__(self, name=name, identifier=identifier)
        self.solution = None
        self.phase = 1
        self.oToken = oToken
        self.action = 'new_line'
        self.configuration.append('action')
        self.preserve_comment = False
        self.insert_whitespace = False
        self.configuration_documentation_link = 'configuring_move_token_rules_link'

    def _get_tokens_of_interest(self, oFile):
        if self.action == 'new_line' and not self.preserve_comment:
            return get_toi_for_new_line_option(self, oFile)
        elif self.action == 'new_line' and self.preserve_comment:
            return get_toi_for_new_line_option_with_preserve_comment(self, oFile)
        else:
            return get_toi_for_move_left_option(self, oFile)

    def _analyze(self, lToi):
        if self.action == 'new_line' and not self.preserve_comment:
            analyze_new_line(self, lToi)
        elif self.action == 'new_line' and self.preserve_comment:
            analyze_new_line_with_preserve_comment(self, lToi)
        else:
            analyze_move_left(self, lToi)

    def _fix_violation(self, oViolation):
        if self.action == 'new_line' and not self.preserve_comment:
            fix_new_line_violations(oViolation)
        elif self.action == 'new_line' and self.preserve_comment:
            fix_new_line_with_preserve_comment_violations(oViolation)
        else:
            fix_move_left_violations(oViolation)


def tokens_are_next_to_each_other(lTokens):
    if len(lTokens) == 2:
        return True
    return False


def get_toi_for_new_line_option(self, oFile):
    lReturn = []
    lToi = oFile.get_token_and_n_tokens_before_it([self.oToken], 2)
    for oToi in lToi:
        lTokens = oToi.get_tokens()
        if isinstance(lTokens[0], parser.carriage_return) or isinstance(lTokens[1], parser.carriage_return):
            continue
        lReturn.append(oToi)
    return lReturn


def get_toi_for_new_line_option_with_preserve_comment(self, oFile):
    lReturn = []
    lToi = oFile.get_line_which_includes_tokens([self.oToken])
    for oToi in lToi:
        lTokens = oToi.get_tokens()
        if not rules_utils.token_at_the_beginning_of_a_line(self.oToken, lTokens):
            lReturn.append(oToi)
    return lReturn


def get_toi_for_move_left_option(self, oFile):
    lReturn = []
    lToi = oFile.get_tokens_between_non_whitespace_token_and_token(self.oToken)
    for oToi in lToi:
        if not tokens_are_next_to_each_other(oToi.get_tokens()):
            lReturn.append(oToi)
    return lReturn


def analyze_new_line(self, lToi):
    for oToi in lToi:
        lTokens = oToi.get_tokens()
        sSolution = f'Move {lTokens[-1].get_value()} to next line.'
        oViolation = violation.New(oToi.get_line_number(), oToi, sSolution)
        self.add_violation(oViolation)


def analyze_new_line_with_preserve_comment(self, lToi):
    for oToi in lToi:
        lTokens = oToi.get_tokens()
        iTokenIndex = oToi.token_index
        sSolution = f'Move {lTokens[iTokenIndex].get_value()} to next line.'
        oViolation = violation.New(oToi.get_line_number(), oToi, sSolution)
        self.add_violation(oViolation)


def analyze_move_left(self, lToi):
    for oToi in lToi:
        lTokens = oToi.get_tokens()
        if rules_utils.number_of_carriage_returns(lTokens) > 0:
            oViolation = create_move_left_violation(self, oToi)
            self.add_violation(oViolation)


def create_move_left_violation(self, oToi):
    lTokens = oToi.get_tokens()
    oFirstToken = lTokens[0]
    oLastToken = lTokens[-1]
    sSolution = f'Move {oLastToken.get_value()} next to {oFirstToken.get_value()}.'
    oViolation = violation.New(oToi.get_line_number(), oToi, sSolution)
    oViolation.insert_whitespace = False
    oViolation.insert_whitespace = self.insert_whitespace
    return oViolation


def fix_new_line_violations(oViolation):
    lTokens = oViolation.get_tokens()
    if isinstance(lTokens[1], parser.whitespace):
        rules_utils.insert_carriage_return(lTokens, -2)
    else:
        rules_utils.insert_carriage_return(lTokens, -1)
    oViolation.set_tokens(lTokens)


def fix_new_line_with_preserve_comment_violations(oViolation):
    lTokens = oViolation.get_tokens()
    iTokenIndex = oViolation.oTokens.token_index
    lTemp = []
    if isinstance(lTokens[-1], parser.comment):
        lTemp.append(lTokens.pop())
        if isinstance(lTokens[-1], parser.whitespace):
            lTemp.insert(0, lTokens.pop())
    rules_utils.insert_carriage_return(lTokens, iTokenIndex)
    lTokens[iTokenIndex:iTokenIndex] = lTemp

    oViolation.set_tokens(lTokens)


def fix_move_left_violations(oViolation):
    lTokens = oViolation.get_tokens()

    rules_utils.insert_token(lTokens, 1, lTokens.pop())
    if oViolation.insert_whitespace:
        rules_utils.insert_token(lTokens, 1, parser.whitespace(' '))

    lNewTokens = utils.remove_consecutive_whitespace_tokens(lTokens)
    lNewTokens = utils.fix_blank_lines(lNewTokens)

    oViolation.set_tokens(lNewTokens)




from vsg.rules import token_indent

from vsg import token

lTokens = []
lTokens.append(token.wait_statement.wait_keyword)
lTokens.append(token.wait_statement.label)


class rule_001(token_indent):
    '''
    This rule checks for indentation of the **wait** keyword.
    Proper indentation enhances comprehension.

    **Violation**

    .. code-block:: vhdl

       begin

           wait for 10ns;
        wait on a,b;
              wait until a = '0';

    **Fix**

    .. code-block:: vhdl

       begin

         wait for 10ns;
         wait on a,b;
         wait until a = '0';
    '''

    def __init__(self):
        token_indent.__init__(self, 'wait', '001', lTokens)




from vsg.vhdlFile.classify import configuration_declarative_item


def detect(iToken, lObjects):
    '''
    configuration_declarative_part ::=
        { configuration_declarative_item }
    '''

    iLast = 0
    iCurrent = iToken
    while iLast != iCurrent:
        iLast = iCurrent
        iCurrent = configuration_declarative_item.detect(iCurrent, lObjects)
    return iCurrent




from vsg.vhdlFile.classify import context_reference
from vsg.vhdlFile.classify import library_clause
from vsg.vhdlFile.classify import use_clause


def detect(iToken, lObjects):
    '''
    context_item ::=
        library_clause
      | use_clause
      | context_reference
    '''

    iCurrent = library_clause.detect(iToken, lObjects)
    if iCurrent != iToken:
        return iCurrent

    iCurrent = use_clause.detect(iToken, lObjects)
    if iCurrent != iToken:
        return iCurrent

    iCurrent = context_reference.detect(iToken, lObjects)
    if iCurrent != iToken:
        return iCurrent

    return iToken



from vsg import parser


class procedure_keyword(parser.keyword):
    '''
    unique_id = subprogram_kind : procedure_keyword
    '''

    def __init__(self, sString):
        parser.keyword.__init__(self, sString)


class function_keyword(parser.keyword):
    '''
    unique_id = subprogram_kind : function_keyword
    '''

    def __init__(self, sString):
        parser.keyword.__init__(self, sString)




from vsg import parser


class assignment(parser.assignment):
    '''
    unique_id = element_association : assignment
    '''

    def __init__(self, sString):
        parser.assignment.__init__(self, sString)




from vsg.vhdlFile.classify import protected_type_declaration
from vsg.vhdlFile.classify import protected_type_body


def detect(iToken, lObjects):
    '''
    protected_type_definition ::=
        protected_type_declaration
      | protected_type_body
    '''

    iReturn = protected_type_declaration.detect(iToken, lObjects)
    if iReturn != iToken:
        return iReturn

    iReturn = protected_type_body.detect(iToken, lObjects)
    if iReturn != iToken:
        return iReturn

    return iToken




from vsg.rules import previous_line

from vsg import token

lTokens = []
lTokens.append(token.package_declaration.package_keyword)


class rule_003(previous_line):
    '''
    This rule checks for blank lines or comments above the **package** keyword.

    |configuring_previous_line_rules_link|

    The default style is :code:`no_code`.

    **Violation**

    .. code-block:: vhdl

       library ieee;
       package FIFO_PKG is

    **Fix**

    .. code-block:: vhdl

       library ieee;

       package FIFO_PKG is
    '''

    def __init__(self):
        previous_line.__init__(self, 'package', '003', lTokens)
        self.style = 'no_code'




from vsg import parser


class open_bracket(parser.open_bracket):
    '''
    unique_id = signature : open_bracket
    '''

    def __init__(self, sString='['):
        parser.open_bracket.__init__(self, '[')


class return_keyword(parser.keyword):
    '''
    unique_id = signature : return_keyword
    '''

    def __init__(self, sString):
        parser.keyword.__init__(self, sString)


class comma(parser.comma):
    '''
    unique_id = signature : comma
    '''

    def __init__(self, sString=','):
        parser.comma.__init__(self)


class close_bracket(parser.close_bracket):
    '''
    unique_id = signature : close_bracket
    '''

    def __init__(self, sString=']'):
        parser.close_bracket.__init__(self, ']')




from vsg import token

from vsg.rules.whitespace_before_tokens_in_between_tokens import Rule

lTokens = []
lTokens.append(token.interface_constant_declaration.colon)
lTokens.append(token.interface_signal_declaration.colon)
lTokens.append(token.interface_variable_declaration.colon)
lTokens.append(token.interface_file_declaration.colon)
lTokens.append(token.interface_unknown_declaration.colon)

oStart = token.port_clause.open_parenthesis
oEnd = token.port_clause.close_parenthesis


class rule_020(Rule):
    '''
    This rule checks for at least one space before the colon.

    |configuring_whitespace_rules_link|

    **Violation**

    .. code-block:: vhdl

       port (
         WR_EN    : in    std_logic;
         RD_EN    : in    std_logic;
         OVERFLOW: out   std_logic;
         DATA     : inout std_logic
       );

    **Fix**

    .. code-block:: vhdl

       port (
         WR_EN    : in    std_logic;
         RD_EN    : in    std_logic;
         OVERFLOW : out   std_logic;
         DATA     : inout std_logic
       );
    '''
    def __init__(self):
        Rule.__init__(self, 'port', '020', lTokens, oStart, oEnd)
        self.number_of_spaces = '>=1'




from vsg.rules import token_indent

from vsg import token

lTokens = []
lTokens.append(token.exit_statement.label)
lTokens.append(token.exit_statement.exit_keyword)


class rule_300(token_indent):
    '''
    This rule checks the indent of the **exit** keyword.

    **Violation**

    .. code-block:: vhdl

       end if;

         exit;

    **Fix**

    .. code-block:: vhdl

       end if;

       exit;
    '''

    def __init__(self):
        token_indent.__init__(self, 'exit_statement', '300', lTokens)




from vsg.rules import remove_comments_from_end_of_lines_bounded_by_tokens

from vsg import token

oStart = token.component_declaration.component_keyword

oEnd = token.component_declaration.semicolon


class rule_019(remove_comments_from_end_of_lines_bounded_by_tokens):
    '''
    This rule checks for comments at the end of the port and generic clauses in component declarations.
    These comments represent additional maintainence.
    They will be out of sync with the entity at some point.
    Refer to the entity for port types, port directions and purpose.

    **Violation**

    .. code-block:: vhdl

       wr_en : in    std_logic;  -- Enables write to RAM
       rd_en : out   std_logic; -- Enable reads from RAM

    **Fix**

    .. code-block:: vhdl

       wr_en : in    std_logic;
       rd_en : out   std_logic;
    '''

    def __init__(self):
        remove_comments_from_end_of_lines_bounded_by_tokens.__init__(self, 'component', '019', oStart, oEnd)
        self.solution = 'Remove comment.'




from vsg import token

from vsg.rules import n_spaces_before_and_after_tokens

lTokens = []
lTokens.append(token.logical_operator.logical_operator)


class rule_013(n_spaces_before_and_after_tokens):
    '''
    This rule checks for at least a single space before and after logical operators.

    **Violation**

    .. code-block:: vhdl

      if (a = '1')and(b = '0')
      if (a = '0')or (b = '1')

    **Fix**

    .. code-block:: vhdl

      if (a = '1') and (b = '0')
      if (a = '0') or (b = '1')
    '''

    def __init__(self):
        n_spaces_before_and_after_tokens.__init__(self, 'whitespace', '013', 1, lTokens, bNIsMinimum=True)
        self.solution = 'Ensure a single space before and after concat operator.'
        self.configuration_documentation_link = None




from vsg.rules import token_prefix as Rule

from vsg.rules import utils

from vsg import token

lTokens = []


class rule_600(Rule):
    '''
    This rule checks for valid prefixes on port identifiers for input ports.

    The default prefix is: *i\_*.

    |configuring_prefix_and_suffix_rules_link|

    **Violation**

    .. code-block:: vhdl

       port (
         wr_en    : in    std_logic;
         rd_en    : in    std_logic
       );


    **Fix**

    .. code-block:: vhdl

       port (
         i_wr_en    : in    std_logic;
         i_rd_en    : in    std_logic
       );
    '''

    def __init__(self):
        Rule.__init__(self, 'port', '600', lTokens)
        self.prefixes = ['i_']

    def _get_tokens_of_interest(self, oFile):
        lReturn = []
        lToi = oFile.get_interface_elements_between_tokens(token.port_clause.open_parenthesis, token.port_clause.close_parenthesis)
        lReturn = utils.extract_identifiers_with_mode_of_input(lToi)
        return lReturn




from vsg import token

from vsg.rules import utils as rules_utils
from vsg.rules import check
from vsg.rules import create_violation
from vsg.rules import fix
from vsg.rules import tokens_of_interest as toi

from vsg.rule_group import structure

from vsg.vhdlFile import utils


class multiline_constraint_structure(structure.Rule):
    '''
    This rule checks the structure of multiline constraints.
    '''

    def __init__(self, name, identifier):
        structure.Rule.__init__(self, name=name, identifier=identifier)
        self.phase = 1
        self.lTokenPairs = None
        self.configuration_documentation_link = 'configuring_multiline_constraint_rules_link'

        self.record_constraint_open_paren = 'ignore'
        self.configuration.append('record_constraint_open_paren')
        self.record_constraint_close_paren = 'ignore'
        self.configuration.append('record_constraint_close_paren')
        self.record_constraint_comma = 'ignore'
        self.configuration.append('record_constraint_comma')
        self.record_constraint_element = 'ignore'
        self.configuration.append('record_constraint_element')
        self.array_constraint = 'ignore'
        self.configuration.append('array_constraint')
        self.exceptions = []
        self.configuration.append('exceptions')

    def _get_tokens_of_interest(self, oFile):
        return toi.get_tokens_bounded_by(self.lTokenPairs, oFile)

    def _analyze(self, lToi):
        for oToi in lToi:

            if exception_applied(self, oToi):
                continue

            _check_record_constraint_open_paren(self, oToi)
            _check_record_constraint_close_paren(self, oToi)
            _check_record_constraint_comma(self, oToi)
            _check_record_constraint_element(self, oToi)
            _check_array_constraint(self, oToi)

        self._sort_violations()

    def _fix_violation(self, oViolation):
        fix.fix_violation(oViolation)


def _check_array_constraint(self, oToi):

    if self.array_constraint == 'ignore':
        return
    elif self.array_constraint == 'all_in_one_line':
        analyze_array_constraint_with_function(self, oToi, analyze_for_array_constraint_all_in_one_line)
    elif self.array_constraint == 'one_line_per_dimension':
        analyze_array_constraint_with_function(self, oToi, analyze_for_array_constraint_one_line_per_dimension)


def analyze_for_array_constraint_all_in_one_line(self, oToi):
    iStart = oToi.get_meta_data('iStart')
    iToken = oToi.get_meta_data('iToken')
    lTokens = oToi.get_tokens()

    if rules_utils.token_at_beginning_of_line_in_token_list(iStart, lTokens):
        oViolation = create_array_constraint_all_in_one_line_violation(oToi)
        self.add_violation(oViolation)
    elif rules_utils.number_of_carriage_returns(lTokens[iStart:iToken]) > 0:
        oViolation = create_array_constraint_remove_carriage_return_violation(oToi)
        self.add_violation(oViolation)


def create_array_constraint_all_in_one_line_violation(oToi):
    oToi.set_meta_data('sSolution', 'Move open parenthesis to previous line and remove carriage returns in array constraint.')
    oToi.set_meta_data('sAction', 'remove_new_line')
    toi.adjust_start_index_based_on_whitespace(oToi, -2)
    oViolation = create_violation._create_violation(oToi)
    return oViolation


def analyze_array_constraint_with_function(self, oToi, fFunction):
    iLine, lTokens = rules_utils.get_toi_parameters(oToi)
    oStartToken = token.index_constraint.open_parenthesis
    oEndToken = token.index_constraint.close_parenthesis

    for iToken, oToken in enumerate(lTokens):
        iLine = utils.increment_line_number(iLine, oToken)
        if isinstance(oToken, oStartToken):
            oToi.set_meta_data('iStart', iToken)
            oToi.set_meta_data('iStartLine', iLine)
        if isinstance(oToken, oEndToken):
            oToi.set_meta_data('iToken', iToken)
            fFunction(self, oToi)


def analyze_for_array_constraint_one_line_per_dimension(self, oToi):
    iStart = oToi.get_meta_data('iStart')
    iToken = oToi.get_meta_data('iToken')
    lTokens = oToi.get_tokens()
    if not rules_utils.token_at_beginning_of_line_in_token_list(iStart, lTokens):
        oViolation = create_array_constraint_one_line_violation(oToi)
        self.add_violation(oViolation)
    elif rules_utils.number_of_carriage_returns(lTokens[iStart:iToken]) > 0:
        oViolation = create_array_constraint_remove_carriage_return_violation(oToi)
        self.add_violation(oViolation)


def create_array_constraint_one_line_violation(oToi):
    iStart = oToi.get_meta_data('iStart')
    lTokens = oToi.get_tokens()
    oToi.set_meta_data('sSolution', 'Move open parenthesis to next line and remove carriage returns in array constraint.')
    oToi.set_meta_data('sAction', 'add_new_line_and_remove_carraige_returns')
    toi.adjust_start_index_based_on_whitespace(oToi, -1)
    oViolation = create_violation._create_violation(oToi)
    return oViolation


def create_array_constraint_remove_carriage_return_violation(oToi):
    oToi.set_meta_data('sSolution', 'Remove carriage returns in array constraint.')
    oToi.set_meta_data('sAction', 'remove_new_line')
    oViolation = create_violation._create_violation(oToi)
    return oViolation


def _check_record_constraint_open_paren(self, oToi):

    check.add_new_line_and_remove_new_line(self, oToi, self.record_constraint_open_paren, token.record_constraint.open_parenthesis)


def _check_record_constraint_close_paren(self, oToi):

    check.add_new_line_and_remove_new_line(self, oToi, self.record_constraint_close_paren, token.record_constraint.close_parenthesis)


def _check_record_constraint_comma(self, oToi):

    check.add_new_line_and_remove_new_line(self, oToi, self.record_constraint_comma, token.record_constraint.comma)


def _check_record_constraint_element(self, oToi):

    check.add_new_line_and_remove_new_line(self, oToi, self.record_constraint_element, token.record_element_constraint.record_element_simple_name)


def exception_applied(self, oToi):
    if _check_for_exception_one(self, oToi):
        return True
    return False


def _check_for_exception_one(self, oToi):
    if not exception_enabled(self):
        return False
    if not token_pattern_match(oToi):
        return False
    analyze_exception_one(self, oToi)
    return True


def exception_enabled(self):
    if 'keep_record_constraint_with_single_element_on_one_line' in self.exceptions:
        return True
    return False


def analyze_exception_one(self, oToi):
    iLine, lTokens = rules_utils.get_toi_parameters(oToi)
    for iToken, oToken in enumerate(lTokens):
        iLine = utils.increment_line_number(iLine, oToken)
        oToi.set_meta_data('iToken', iToken)
        record_constraint_open_paren_detected(oToi, iLine, oToken)
        record_constraint_close_paren_detected(self, oToi, oToken)


def record_constraint_open_paren_detected(oToi, iLine, oToken):
    if isinstance(oToken, token.record_constraint.open_parenthesis):
        oToi.set_meta_data('iStart', oToi.get_meta_data('iToken'))
        oToi.set_meta_data('iStartLine', iLine)


def record_constraint_close_paren_detected(self, oToi, oToken):
    if isinstance(oToken, token.record_constraint.close_parenthesis):
        lTokens = oToi.get_tokens()
        iStart = oToi.get_meta_data('iStart')
        iToken = oToi.get_meta_data('iToken')
        if rules_utils.number_of_carriage_returns(lTokens[iStart:iToken]) > 0:
            oViolation = create_array_constraint_remove_carriage_return_violation(oToi)
            self.add_violation(oViolation)


def token_pattern_match(oToi):
    lActual = filter_tokens(oToi)
    lExpected = create_expected_token_pattern()
    if len(lExpected) != len(lActual):
        return False
    for iIndex in range(0, len(lExpected)):
        if not isinstance(lActual[iIndex], lExpected[iIndex]):
            return False
    return True


def create_expected_token_pattern():
    lCheckType = []
    lCheckType.append(token.record_constraint.open_parenthesis)
    lCheckType.append(token.index_constraint.open_parenthesis)
    lCheckType.append(token.index_constraint.close_parenthesis)
    lCheckType.append(token.record_constraint.close_parenthesis)
    return lCheckType


def filter_tokens(oToi):
    lReturn = []
    lTokens = oToi.get_tokens()
    for oToken in lTokens:
        add_index_constraint_open_parenthesis(oToken, lReturn)
        add_index_constraint_close_parenthesis(oToken, lReturn)
        add_record_constraint_open_parenthesis(oToken, lReturn)
        add_record_constraint_close_parenthesis(oToken, lReturn)
    return lReturn


def add_index_constraint_open_parenthesis(oToken, lReturn):
    add_token_type(oToken, lReturn, token.index_constraint.open_parenthesis)


def add_index_constraint_close_parenthesis(oToken, lReturn):
    add_token_type(oToken, lReturn, token.index_constraint.close_parenthesis)


def add_record_constraint_open_parenthesis(oToken, lReturn):
    add_token_type(oToken, lReturn, token.record_constraint.open_parenthesis)


def add_record_constraint_close_parenthesis(oToken, lReturn):
    add_token_type(oToken, lReturn, token.record_constraint.close_parenthesis)


def add_token_type(oToken, lReturn, oTokenType):
    if isinstance(oToken, oTokenType):
        lReturn.append(oTokenType())




import copy

from vsg import parser
from vsg import token
from vsg import violation

from vsg.rule_group import structure
from vsg.vhdlFile import utils


class separate_multiple_signal_identifiers_into_individual_statements(structure.Rule):
    '''
    Checks the case for words.

    Parameters
    ----------

    name : string
       The group the rule belongs to.

    identifier : string
       unique identifier.  Usually in the form of 00N.

    lTokens : list of parser object types
       object types to check the prefix

    lPrefixes : string list
       acceptable prefixes
    '''

    def __init__(self, name, identifier, lTokens, iAllow=2):
        structure.Rule.__init__(self, name=name, identifier=identifier)
        self.solution = 'Split signal declaration into individual declarations'
        self.lTokens = lTokens
        self.consecutive = iAllow
        self.configuration.append('consecutive')
        self.configuration_documentation_link = 'configuring_number_of_signals_in_signal_declaration_link'

    def _get_tokens_of_interest(self, oFile):
        return oFile.get_tokens_bounded_by(token.signal_declaration.signal_keyword, token.signal_declaration.semicolon)

    def _analyze(self, lToi):
        for oToi in lToi:
            lTokens = oToi.get_tokens()
            iIdentifiers = 0
            iStartIndex = None
            iEndIndex = 0
            bPreTokens = True
            lPreTokens = []
            lIdentifiers = []
            for iToken, oToken in enumerate(lTokens):
                if isinstance(oToken, token.signal_declaration.identifier):
                    lIdentifiers.append(oToken)
                    iIdentifiers += 1
                    if iStartIndex is None:
                        iStartIndex = iToken
                    iEndIndex = iToken
                    bPreTokens = False

                if bPreTokens:
                    lPreTokens.append(oToken)

            if iIdentifiers > self.consecutive:
                oViolation = violation.New(oToi.get_line_number(), oToi, self.solution)
                dAction = {}
                dAction['start'] = iStartIndex
                dAction['end'] = iEndIndex
                dAction['number'] = iIdentifiers
                dAction['identifiers'] = lIdentifiers
                oViolation.set_action(dAction)
                self.add_violation(oViolation)

    def _fix_violation(self, oViolation):
        lTokens = oViolation.get_tokens()
        dAction = oViolation.get_action()

        lFinalTokens = []
        for oIdentifier in dAction['identifiers']:
            lNewTokens = []
            for iToken, oToken in enumerate(lTokens):
                if iToken < dAction['start']:
                    lNewTokens.append(copy.deepcopy(oToken))
                if iToken == dAction['start']:
                    lNewTokens.append(oIdentifier)
                if iToken > dAction['end']:
                    lNewTokens.append(copy.deepcopy(oToken))
            lNewTokens = utils.remove_carriage_returns_from_token_list(lNewTokens)
            lFinalTokens.extend(lNewTokens)
            lFinalTokens.append(parser.carriage_return())

        lFinalTokens.pop()
        oViolation.set_tokens(lFinalTokens)




from vsg import parser
from vsg import token
from vsg import violation

from vsg.rules import utils as rules_utils
from vsg.rule_group import structure
from vsg.vhdlFile import utils


class multiline_simple_structure(structure.Rule):
    '''
    This rule checks the structure of simple and conditional concurrent statements.

    |configuring_multiline_structure_rules_link|

    **Violation**

    .. code-block:: vhdl

       wr_en <=
         '0' when q_wr_en = '1' else
                '1';

       w_foo <=
         I_FOO when ((I_BAR = '1') and
                            (I_CRUFT = '1')) else
                '0';

    **Fix**

    .. code-block:: vhdl

       wr_en <= '0' when q_wr_en = '1' else
                '1';

       w_foo <= I_FOO when ((I_BAR = '1') and
                            (I_CRUFT = '1')) else
                '0';
    '''

    def __init__(self, name, identifier, lTokenPairs):
        structure.Rule.__init__(self, name=name, identifier=identifier)
        self.lTokenPairs = lTokenPairs

        self.new_line_after_assign = 'no'
        self.configuration.append('new_line_after_assign')
        self.ignore_single_line = 'yes'
        self.configuration.append('ignore_single_line')

    def _get_tokens_of_interest(self, oFile):
        lToi = []
        for lTokenPair in self.lTokenPairs:
            aToi = oFile.get_tokens_bounded_by(lTokenPair[0], lTokenPair[1])
            lToi = utils.combine_two_token_class_lists(lToi, aToi)

        lReturn = []
        for oToi in lToi:
            if toi_is_an_array(oToi):
                continue
            lReturn.append(oToi)

        return lReturn

    def _analyze(self, lToi):
        for oToi in lToi:
            if rules_utils.is_single_line(oToi) and self.ignore_single_line:
                continue

            _check_new_line_after_assign(self, oToi)

        self._sort_violations()

    def _fix_violation(self, oViolation):
        dAction = oViolation.get_action()
        if dAction['type'] == 'new_line_after_assign':
            _fix_new_line_after_assign(oViolation)


def _check_new_line_after_assign(self, oToi):

    if self.new_line_after_assign == 'ignore':
        return

    iLine, lTokens = utils.get_toi_parameters(oToi)

    iNextToken = utils.find_next_non_whitespace_token(1, lTokens)

    iNumCarriageReturns = utils.count_carriage_returns(lTokens[:iNextToken])

    if iNumCarriageReturns == 0:
        if self.new_line_after_assign == 'yes':
            sSolution = 'Add return after assignment.'
            oViolation = _create_violation(oToi, iLine, 1, 1, 'new_line_after_assign', 'insert', sSolution)
            self.add_violation(oViolation)
    else:
        if self.new_line_after_assign == 'no':
            sSolution = 'Move code after assignment to the same line as assignment.'
            oViolation = _create_violation(oToi, iLine, 0, iNextToken, 'new_line_after_assign', 'remove', sSolution)
            self.add_violation(oViolation)


def _fix_new_line_after_assign(oViolation):
    lTokens = oViolation.get_tokens()
    dAction = oViolation.get_action()
    if dAction['action'] == 'insert':
        if not isinstance(lTokens[0], parser.whitespace):
            rules_utils.insert_whitespace(lTokens, 0)
        rules_utils.insert_carriage_return(lTokens, 0)
        oViolation.set_tokens(lTokens)
    elif dAction['action'] == 'remove':
        lNewTokens = []
        lNewTokens.append(lTokens[0])
        rules_utils.append_whitespace(lNewTokens)
        lNewTokens.append(lTokens[-1])
        oViolation.set_tokens(lNewTokens)


def _create_violation(oToi, iLine, iStartIndex, iEndIndex, sType, sAction, sSolution):
    dAction = _create_action_dictionary(sType, sAction)
    oViolation = violation.New(iLine, oToi.extract_tokens(iStartIndex, iEndIndex), sSolution)
    oViolation.set_action(dAction)
    return oViolation


def _create_action_dictionary(sType, sAction):
    dReturn = {}
    dReturn['type'] = sType
    dReturn['action'] = sAction
    return dReturn


def toi_is_an_array(oToi):
    lTokens = oToi.get_tokens()
    for oToken in lTokens:
        if isinstance(oToken, token.aggregate.open_parenthesis):
            return True
    return False




from vsg import parser


class shift_operator(parser.keyword):
    '''
    unique_id = shift_operator : shift_operator
    '''

    def __init__(self, sString):
        parser.keyword.__init__(self, sString)


class sll(shift_operator):
    '''
    unique_id = shift_operator : sll
    '''

    def __init__(self, sString):
        shift_operator.__init__(self, sString)


class srl(shift_operator):
    '''
    unique_id = shift_operator : srl
    '''

    def __init__(self, sString):
        shift_operator.__init__(self, sString)


class sla(shift_operator):
    '''
    unique_id = shift_operator : sla
    '''

    def __init__(self, sString):
        shift_operator.__init__(self, sString)


class sra(shift_operator):
    '''
    unique_id = shift_operator : sra
    '''

    def __init__(self, sString):
        shift_operator.__init__(self, sString)


class rol(shift_operator):
    '''
    unique_id = shift_operator : rol
    '''

    def __init__(self, sString):
        shift_operator.__init__(self, sString)


class ror(shift_operator):
    '''
    unique_id = shift_operator : ror
    '''

    def __init__(self, sString):
        shift_operator.__init__(self, sString)




from vsg.token import iteration_scheme as token

from vsg.vhdlFile import utils

from vsg.vhdlFile.classify import condition
from vsg.vhdlFile.classify import parameter_specification


def detect(iToken, lObjects):
    '''
    iteration_scheme ::=
        while condition
      | for *loop*_parameter_specification
    '''
    if utils.find_in_next_n_tokens(';', 3, iToken, lObjects):
        return False
    if utils.find_in_next_n_tokens('else', 3, iToken, lObjects):
        return False
    if utils.find_in_next_n_tokens('while', 3, iToken, lObjects):
        return True
    if utils.find_in_next_n_tokens('for', 3, iToken, lObjects):
        return True
    return False


def classify(iToken, lObjects):

    if utils.is_next_token('while', iToken, lObjects):
        iCurrent = utils.assign_next_token_required('while', token.while_keyword, iToken, lObjects)
        iCurrent = condition.classify_until(['loop'], iToken, lObjects)
        return iCurrent

    if utils.is_next_token('for', iToken, lObjects):
        iCurrent = utils.assign_next_token_required('for', token.for_keyword, iToken, lObjects)
        iCurrent = parameter_specification.classify_until(['loop'], iToken, lObjects)
        return iCurrent

    return iToken




from vsg import parser
from vsg import token
from vsg import violation

from vsg.rule_group import structure

lTokens = []
lTokens.append(token.interface_constant_declaration.assignment)
lTokens.append(token.interface_signal_declaration.assignment)
lTokens.append(token.interface_variable_declaration.assignment)
lTokens.append(token.interface_unknown_declaration.assignment)


class rule_012(structure.Rule):
    '''
    This rule checks for default assignments on port declarations.

    This rule is defaulted to not fixable and can be overridden with a configuration to remove the default assignments.

    **Violation**

    .. code-block:: vhdl

       port (
         I_WR_EN    : in    std_logic := '0';
         I_RD_EN    : in    std_logic := '0';
         O_OVERFLOW : out   std_logic;
         IO_DATA    : inout std_logic := (others => 'Z')
       );

    **Fix**

    .. code-block:: vhdl

       port (
         I_WR_EN    : in    std_logic;
         I_RD_EN    : in    std_logic;
         O_OVERFLOW : out   std_logic;
         IO_DATA    : inout std_logic
       );
    '''

    def __init__(self):
        structure.Rule.__init__(self, name='port', identifier='012')
        self.solution = 'Remove assignment'
        self.lTokens = lTokens
        self.oStart = token.port_clause.open_parenthesis
        self.oEnd = token.port_clause.close_parenthesis
        self.fixable = False
        self.configuration_documentation_link = None

    def _get_tokens_of_interest(self, oFile):
        return oFile.get_interface_elements_between_tokens(self.oStart, self.oEnd)

    def _analyze(self, lToi):
        for oToi in lToi:
            lTokens = oToi.get_tokens()
            for iToken, oToken in enumerate(lTokens):
                for oSearchToken in self.lTokens:
                    if isinstance(oToken, oSearchToken):
                        oViolation = violation.New(oToi.get_line_number(), oToi, self.solution)
                        dAction = {}
                        if isinstance(lTokens[iToken - 1], parser.whitespace):
                            dAction['index'] = iToken - 1
                        else:
                            dAction['index'] = iToken
                        oViolation.set_action(dAction)
                        self.add_violation(oViolation)
                        break

    def _fix_violation(self, oViolation):
        lTokens = oViolation.get_tokens()
        dAction = oViolation.get_action()
        for iIndex in range(dAction['index'], len(lTokens)):
            if isinstance(lTokens[iIndex], parser.comment):
                if isinstance(lTokens[iIndex - 1], parser.whitespace):
                    iEnd = iIndex - 1
                else:
                    iEnd = iIndex
                break
        else:
            iEnd = iIndex
        if iEnd + 1 == len(lTokens):
            lNewTokens = lTokens[:dAction['index']]
        else:
            lNewTokens = lTokens[:dAction['index']] + lTokens[iEnd:]
        oViolation.set_tokens(lNewTokens)




from vsg import token

from vsg.rules.whitespace_between_token_pairs import Rule

lTokens = []
lTokens.append([token.process_statement.process_keyword, token.process_statement.open_parenthesis])
lTokens.append([token.process_statement.process_keyword, token.process_statement.is_keyword])
lTokens.append([token.process_statement.process_keyword, token.process_statement.begin_keyword])


class rule_002(Rule):
    '''
    This rule checks for a single space after the **process** keyword.

    |configuring_whitespace_rules_link|

    **Violation**

    .. code-block:: vhdl

       proc_a : process(rd_en, wr_en, data_in, data_out,

       proc_a : process    (rd_en, wr_en, data_in, data_out,

    **Fix**

    .. code-block:: vhdl

       proc_a : process (rd_en, wr_en, data_in, data_out,

       proc_a : process (rd_en, wr_en, data_in, data_out,
    '''
    def __init__(self):
        Rule.__init__(self, 'process', '002', lTokens)




from vsg import deprecated_rule


class rule_003(deprecated_rule.Rule):
    '''
    This rule has been merged into `function_100 <function_rules.html#function-100>`_.
    '''

    def __init__(self):
        deprecated_rule.Rule.__init__(self, 'function', '003')
        self.message.append('Rule ' + self.unique_id + ' has been merged into function_100.')




from vsg.rules import token_indent

from vsg import token

lTokens = []
lTokens.append(token.enumeration_type_definition.enumeration_literal)


class rule_005(token_indent):
    '''
    This rule checks the indent of multiline enumerated types.

    **Violation**

    .. code-block:: vhdl

       type state_machine is (
       idle,
         write,
       read,
          done);

    **Fix**

    .. code-block:: vhdl

       type state_machine is (
         idle,
         write,
         read,
         done);
    '''

    def __init__(self):
        token_indent.__init__(self, 'type', '005', lTokens)
        self.solution = 'Ensure proper indentation.'




from vsg.token import interface_subprogram_default as token

from vsg.vhdlFile import utils


def classify(iToken, lObjects):
    '''
    interface_subprogram_default ::=
        *subprogram*_name | <>
    '''
    if utils.is_next_token('<>', iToken, lObjects):
        return utils.assign_next_token_required('<>', token.undefined_range, iToken, lObjects)
    return utils.classify_next_token(token.subprogram_name, iToken, lObjects)




from vsg.token import if_statement as token

from vsg.vhdlFile import utils

from vsg.vhdlFile.extract import tokens


def get_if_statement_conditions(lAllTokens, oTokenMap, fRemoveWhitespace=True):
    lReturn = []

    lStart = oTokenMap.get_token_indexes(token.if_keyword, bCopy=True)
    lStart.extend(oTokenMap.get_token_indexes(token.elsif_keyword, bCopy=True))
    lStart.sort()

    lEnd = []
    for iStart in lStart:
        lEnd.append(oTokenMap.get_index_of_token_after_index(token.then_keyword, iStart))

    for iStart, iEnd in zip(lStart, lEnd):
        lTemp = lAllTokens[iStart + 1: iEnd]
        iStartIndex = iStart + 1
        if fRemoveWhitespace:
            iStartIndex, lTemp = utils.remove_leading_whitespace_and_comments(iStart, lTemp)
            lTemp = utils.remove_trailing_whitespace_and_comments(lTemp)

        iLine = oTokenMap.get_line_number_of_index(iStartIndex)

        lReturn.append(tokens.New(iStartIndex, iLine, lTemp))

    return lReturn




from vsg.rules import blank_line_below_line_ending_with_token as Rule

from vsg.token import record_type_definition as token


class rule_200(Rule):
    '''
    This rule checks for blank lines below the **record** keyword.

    |configuring_blank_lines_link|

    **Violation**

    .. code-block:: vhdl

       type t_record is record

         a : std_logic;
         b : std_logic;
       end record    t_record;

    **Fix**

    .. code-block:: vhdl

       type t_record is record
         a : std_logic;
         b : std_logic;
       end record t_record;
    '''

    def __init__(self):
        Rule.__init__(self, 'record_type_definition', '200', [token.record_keyword])
        self.style = 'no_blank_line'




from vsg.vhdlFile.classify import architecture_body
from vsg.vhdlFile.classify import package_body


def detect(iToken, lObjects):
    '''
    secondary_unit ::=
        architecture_body
      | package_body
    '''
    iReturned = architecture_body.detect(iToken, lObjects)
    if iReturned != iToken:
        return iReturned

    iReturned = package_body.detect(iToken, lObjects)
    if iReturned != iToken:
        return iReturned

    return iToken




from vsg import parser


class signal_keyword(parser.keyword):
    '''
    unique_id = interface_signal_declaration : signal_keyword
    '''

    def __init__(self, sString):
        parser.keyword.__init__(self, sString)


class identifier(parser.identifier):
    '''
    unique_id = interface_signal_declaration : identifier
    '''

    def __init__(self, sString):
        parser.identifier.__init__(self, sString)


class comma(parser.comma):
    '''
    unique_id = interface_signal_declaration : comma
    '''

    def __init__(self, sString=','):
        parser.comma.__init__(self)


class colon(parser.colon):
    '''
    unique_id = interface_signal_declaration : colon
    '''

    def __init__(self, sString=':'):
        parser.colon.__init__(self)


class mode_keyword(parser.keyword):
    '''
    unique_id = interface_signal_declaration : mode_keyword
    '''

    def __init__(self, sString):
        parser.keyword.__init__(self, sString)


class subtype_indication(parser.subtype_indication):
    '''
    unique_id = interface_signal_declaration : subtype_indication
    '''

    def __init__(self, sString):
        parser.subtype_indication.__init__(self, sString)


class bus_keyword(parser.keyword):
    '''
    unique_id = interface_signal_declaration : bus_keyword
    '''

    def __init__(self, sString):
        parser.keyword.__init__(self, sString)


class assignment(parser.assignment):
    '''
    unique_id = interface_signal_declaration : assignment
    '''

    def __init__(self, sString):
        parser.assignment.__init__(self, sString)


class static_expression(parser.static_expression):
    '''
    unique_id = interface_signal_declaration : static_expression
    '''

    def __init__(self, sString):
        parser.static_expression.__init__(self, sString)


class semicolon(parser.semicolon):
    '''
    unique_id = interface_signal_declaration : semicolon
    '''

    def __init__(self):
        parser.semicolon.__init__(self)




from vsg import parser
from vsg import token

from vsg.vhdlFile import utils


def set_token_indent(dIndentMap, lTokens):
    lTokenKeys, dIndents = process_indent_map(dIndentMap)
    cParams = parameters()
    cParams.lTokenKeys = lTokenKeys
    cParams.dIndents = dIndents

    for iToken, oToken in enumerate(lTokens):

        if isinstance(oToken, parser.whitespace):
            continue

        if isinstance(oToken, parser.blank_line):
            cParams.bLibraryFound = False
            continue

        if isinstance(oToken, parser.carriage_return):
            continue

        sUniqueId = oToken.get_unique_id(sJoin=':')

        iTokenIndent = None

        if sUniqueId in cParams.lTokenKeys:
            token_key = cParams.dIndents[sUniqueId]['token']
            after_key = cParams.dIndents[sUniqueId]['after']
            iTokenIndent = update_indent_var(cParams.iIndent, token_key)
            oToken.set_indent(iTokenIndent)
            cParams.iIndent = update_indent_var(cParams.iIndent, after_key)

        if isinstance(oToken, token.context_declaration.end_keyword):
            cParams.bLibraryFound = False
            continue

        if isinstance(oToken, token.library_clause.keyword):
            cParams.bLibraryFound = True
            continue

        if isinstance(oToken, token.use_clause.keyword):
            if not cParams.bArchitectureFound:
#                oToken.set_indent(cParams.iIndent + 1)
                sIndent = cParams.dIndents[sUniqueId]['token_after_library_clause']
                iIndent = update_indent_var(cParams.iIndent, sIndent)
                oToken.set_indent(iIndent)
            else:
                oToken.set_indent(cParams.iIndent)
            continue

        if isinstance(oToken, token.context_reference.keyword):
            if cParams.bLibraryFound:
                oToken.set_indent(cParams.iIndent + 1)
            else:
                oToken.set_indent(cParams.iIndent)
            continue

        if isinstance(oToken, token.architecture_body.architecture_keyword):
            cParams.bLibraryFound = False
            cParams.bArchitectureFound = True
            continue

        if isinstance(oToken, token.architecture_body.semicolon):
            cParams.bArchitectureFound = False
            continue

        if isinstance(oToken, token.entity_declaration.entity_keyword):
            cParams.bLibraryFound = False
            continue

        if isinstance(oToken, token.package_body.package_keyword):
            cParams.bLibraryFound = False
            continue

        if isinstance(oToken, token.package_declaration.package_keyword):
            cParams.bLibraryFound = False
            continue

        ### Comments
        if isinstance(oToken, parser.comment):
            set_indent_of_comment(cParams, iToken, lTokens)
            continue

        ### Concurrent signal assignment
        if isinstance(oToken, token.concurrent_signal_assignment_statement.label_name):
            cParams.insideConcurrentSignalAssignment = True
            continue

        if isinstance(oToken, token.concurrent_signal_assignment_statement.postponed_keyword):
            cParams.insideConcurrentSignalAssignment = True
            continue

        if isinstance(oToken, token.concurrent_simple_signal_assignment.target):
            cParams.insideConcurrentSignalAssignment = True
            continue

        if isinstance(oToken, token.concurrent_conditional_signal_assignment.target):
            cParams.insideConcurrentSignalAssignment = True
            continue

        if isinstance(oToken, token.concurrent_selected_signal_assignment.with_keyword):
            cParams.insideConcurrentSignalAssignment = True
            continue

        if isinstance(oToken, token.concurrent_simple_signal_assignment.semicolon):
            cParams.insideConcurrentSignalAssignment = False
            continue

        if isinstance(oToken, token.concurrent_conditional_signal_assignment.semicolon):
            cParams.insideConcurrentSignalAssignment = False
            continue

        if isinstance(oToken, token.concurrent_selected_signal_assignment.semicolon):
            cParams.insideConcurrentSignalAssignment = False
            continue

        oToken.set_indent(iTokenIndent)


def update_indent_var(iIndent, update):
    if str(update).startswith('+'):
        return iIndent + int(update)
    if str(update).startswith('-'):
        return iIndent + int(update)
    if update == 'current':
        return iIndent
    return update


def process_indent_map(dIndentMap):
    dReturn = {}
    lTopKeys = list(dIndentMap['indent']['tokens'].keys())
    for sTopKey in lTopKeys:
        lSubKeys = list(dIndentMap['indent']['tokens'][sTopKey].keys())
        for sSubKey in lSubKeys:
            sUniqueId = sTopKey + ':' + sSubKey
            dReturn[sUniqueId] = dIndentMap['indent']['tokens'][sTopKey][sSubKey]

    lReturn = list(dReturn.keys())

    return lReturn, dReturn


def is_use_clause_use_keyword_next(iIndex, lTokens):
    iToken = utils.find_next_non_whitespace_token(iIndex, lTokens)
    if isinstance(lTokens[iToken], token.use_clause.keyword):
        return True
    if isinstance(lTokens[iToken], token.context_reference.keyword):
        return True
    return False


def set_indent_of_comment(cParams, iToken, lTokens):
    oToken = lTokens[iToken]
    if cParams.bLibraryFound:
        set_indent_of_use_clause_comment(cParams, iToken, lTokens)
    elif oToken.is_block_comment:
        set_indent_of_block_comment(cParams, iToken, lTokens)
    else:
        set_indent_of_normal_comment(cParams, iToken, lTokens)


def set_indent_of_normal_comment(cParams, iToken, lTokens):
    oToken = lTokens[iToken]
    iTemp = get_indent_value_of_next_token(iToken, lTokens, cParams)
    oToken.set_indent(iTemp)


def set_indent_of_use_clause_comment(cParams, iToken, lTokens):
    oToken = lTokens[iToken]
    if is_use_clause_use_keyword_next(iToken + 1, lTokens):
        oToken.set_indent(cParams.iIndent + 1)
    else:
        oToken.set_indent(cParams.iIndent)


def set_indent_of_block_comment(cParams, iToken, lTokens):
    oToken = lTokens[iToken]
    if oToken.block_comment_indent == 0:
        oToken.set_indent(0)
    else:
        oToken.set_indent(cParams.iIndent)


def get_indent_value_of_next_token(iToken, lTokens, cParams):
    iIndex = utils.find_next_non_whitespace_token(iToken + 1, lTokens)
    sUniqueId = lTokens[iIndex].get_unique_id(sJoin=':')
    if sUniqueId in cParams.lTokenKeys:
        token_key = cParams.dIndents[sUniqueId]['token']
        iTokenIndent = update_indent_var(cParams.iIndent, token_key)
        return iTokenIndent
    return cParams.iIndent


class parameters():

    def __init__(self):
        self.lTokenKeys = None
        self.dIndents = None
        self.bLibraryFound = False
        self.bArchitectureFound = False
        self.insideConcurrentSignalAssignment = False
        self.iIndent = 0




from vsg import token

from vsg.rules.whitespace_before_tokens_in_between_tokens import Rule

lTokens = []
lTokens.append(token.interface_constant_declaration.colon)
lTokens.append(token.interface_signal_declaration.colon)
lTokens.append(token.interface_variable_declaration.colon)
lTokens.append(token.interface_file_declaration.colon)
lTokens.append(token.interface_unknown_declaration.colon)

oStart = token.generic_clause.open_parenthesis
oEnd = token.generic_clause.close_parenthesis


class rule_014(Rule):
    '''
    This rule checks for at least a single space before the colon.

    |configuring_whitespace_rules_link|

    **Violation**

    .. code-block:: vhdl

       g_address_width: integer := 10;
       g_data_width : integer := 32;
       g_depth: integer := 512;

    **Fix**

    .. code-block:: vhdl

       g_address_width : integer := 10;
       g_data_width : integer := 32;
       g_depth : integer := 512;
    '''
    def __init__(self):
        Rule.__init__(self, 'generic', '014', lTokens, oStart, oEnd)
        self.number_of_spaces = '>=1'




from vsg import token

from vsg.rules.whitespace_after_token import Rule

lTokens = []
lTokens.append(token.element_association.assignment)


class rule_101(Rule):
    '''
    This rule checks for a single space after the => in an element_association.

    |configuring_whitespace_rules_link|

    **Violation**

    .. code-block:: vhdl

       a <= (others =>(others =>     '0'));

    **Fix**

    .. code-block:: vhdl

       a <= (others => (others => '0'));
    '''
    def __init__(self):
        Rule.__init__(self, 'element_association', '101', lTokens)




from vsg.rules import blank_line_below_line_ending_with_token

from vsg.token import component_declaration as token


class rule_018(blank_line_below_line_ending_with_token):
    '''
    This rule checks for a blank line below the **end component** line.

    |configuring_blank_lines_link|

    **Violation**

    .. code-block:: vhdl

       end component fifo;
       signal rd_en : std_logic;


    **Fix**

    .. code-block:: vhdl

       end component fifo;

       signal rd_en : std_logic;
    '''

    def __init__(self):
        blank_line_below_line_ending_with_token.__init__(self, 'component', '018', [token.semicolon])




from vsg.rules import multiline_alignment_between_tokens

from vsg import token

lTokenPairs = []
lTokenPairs.append([token.concurrent_selected_signal_assignment.with_keyword, token.concurrent_selected_signal_assignment.semicolon])
lTokenPairs.append([token.selected_force_assignment.with_keyword, token.selected_force_assignment.semicolon])
lTokenPairs.append([token.selected_variable_assignment.with_keyword, token.selected_variable_assignment.semicolon])
lTokenPairs.append([token.selected_waveform_assignment.with_keyword, token.selected_waveform_assignment.semicolon])


class rule_400(multiline_alignment_between_tokens):
    '''
    This rule checks the alignment of multiline selected assignment statements.

    |configuring_multiline_indent_rules_link|

    **Violation**

    .. code-block:: vhdl

       with (mux_select or reset) select addr <=
       "0000" when 0,
               "0001" when 1,
            "1111" when others;

    **Fix**

    .. code-block:: vhdl

       with (mux_select or reset) select addr <=
         "0000" when 0,
         "0001" when 1,
         "1111" when others;
    '''

    def __init__(self):
        multiline_alignment_between_tokens.__init__(self, 'selected_assignment', '400', lTokenPairs)
        self.phase = 5
        self.configuration.remove('align_left')
        self.configuration.remove('align_paren')
        self.align_left = 'yes'
        self.align_paren = 'no'
        self.override = True




from vsg.rules import token_indent

from vsg import token

lTokens = []
lTokens.append(token.package_declaration.package_keyword)


class rule_001(token_indent):
    '''
    This rule checks the indent of the package declaration.

    **Violation**

    .. code-block:: vhdl

       library ieee;

         package FIFO_PKG is

    **Fix**

    .. code-block:: vhdl

       library ieee;

       package FIFO_PKG is
    '''

    def __init__(self):
        token_indent.__init__(self, 'package', '001', lTokens)




from vsg import token

from vsg.rules import remove_carriage_return_after_token as Rule

lTokens = []
lTokens.append(token.selected_expressions.when_keyword)
lTokens.append(token.selected_waveforms.when_keyword)


class rule_011(Rule):
    '''
    This rule checks the choice is on the same line as the **when** keyword.

    **Violation**

    .. code-block:: vhdl

       with
         mux_sel select addr <=
         "0000" when
         0,
         "0001" when 1,
         "1111" when others;

    **Fix**

    .. code-block:: vhdl

       with mux_sel select addr <=
         "0000" when 0,
         "0001" when 1,
         "1111" when others;
    '''

    def __init__(self):
        Rule.__init__(self, 'selected_assignment', '011', lTokens, bInsertSpace=True)
        self.solution = 'Removed carraige returns after when keyword'




from .rule_001 import rule_001
from .rule_002 import rule_002
from .rule_003 import rule_003
from .rule_004 import rule_004
from .rule_005 import rule_005
from .rule_006 import rule_006
from .rule_007 import rule_007
from .rule_008 import rule_008
from .rule_009 import rule_009
from .rule_010 import rule_010
from .rule_011 import rule_011
from .rule_012 import rule_012

from .rule_100 import rule_100
from .rule_101 import rule_101
from .rule_102 import rule_102
from .rule_103 import rule_103
from .rule_104 import rule_104
from .rule_105 import rule_105
from .rule_106 import rule_106
from .rule_107 import rule_107

from .rule_300 import rule_300

from .rule_400 import rule_400

from .rule_500 import rule_500
from .rule_501 import rule_501
from .rule_502 import rule_502
from .rule_503 import rule_503




from vsg.rules import move_token_next_to_another_token as Rule

from vsg.token import record_type_definition as token


class rule_006(Rule):
    '''
    This rule checks the optional simple name is on the same line as the **record** keyword.

    **Violation**

    .. code-block:: vhdl

       type t_record is record
         a : std_logic;
         b : std_logic;
       end record
       t_record;

    **Fix**

    .. code-block:: vhdl

       type t_record is record
         a : std_logic;
         b : std_logic;
       end record t_record
       ;
    '''

    def __init__(self):
        Rule.__init__(self, 'record_type_definition', '006', token.end_record_keyword, token.record_type_simple_name)
        self.subphase = 2
        self.solution = 'Move simple name next to *record* keyword'



from vsg import parser


class subprogram_name(parser.name):
    '''
    unique_id = interface_subprogram_default : subprogram_name
    '''

    def __init__(self, sString):
        parser.name.__init__(self, sString)


class undefined_range(parser.undefined_range):
    '''
    unique_id = interface_subprogram_default : undefined_range
    '''

    def __init__(self, sString='<>'):
        parser.undefined_range.__init__(self)




from vsg.rules import align_consecutive_lines_starting_with_a_comment_above_line_starting_with_token

from vsg.token import use_clause as token


class rule_009(align_consecutive_lines_starting_with_a_comment_above_line_starting_with_token):
    '''
    This rule checks alignment of comments above library use statements.

    **Violation**

    .. code-block:: vhdl

        library ieee;
        -- Use standard logic library
          use ieee.std_logic_1164.all;

    **Fix**

    .. code-block:: vhdl

        library ieee;
          -- Use standard logic library
          use ieee.std_logic_1164.all;
    '''

    def __init__(self):
        align_consecutive_lines_starting_with_a_comment_above_line_starting_with_token.__init__(self, 'library', '009', token.keyword, bIncrement=True)
        self.solution = 'Align comment with *use* keyword.'
        self.subphase = 2
        self.configuration_documentation_link = None




from vsg.token import case_generate_alternative as token

from vsg.vhdlFile import utils

from vsg.vhdlFile.classify import choices
from vsg.vhdlFile.classify import generate_statement_body


def detect(iToken, lObjects):
    '''
    case_generate_alternative ::=
        when [ alternative_label : ] choices =>
            generate_statement_body
    '''

    if utils.is_next_token('when', iToken, lObjects):
        return classify(iToken, lObjects)
    return iToken


def classify(iToken, lObjects):

    iCurrent = utils.assign_next_token_required('when', token.when_keyword, iToken, lObjects)

    iCurrent = choices.classify_until(['=>'], iCurrent, lObjects)

    iCurrent = utils.assign_next_token_required('=>', token.assignment, iToken, lObjects)

    iCurrent = generate_statement_body.classify(iCurrent, lObjects)

    return iCurrent




from vsg import deprecated_rule


class rule_010(deprecated_rule.Rule):
    '''
    The function of this rule has been moved to rule ieee_500.
    '''
    def __init__(self):
        deprecated_rule.Rule.__init__(self, 'signal', '010')
        self.message.append('Rule ' + self.unique_id + ' has been replaced with rule ieee_500.')




from vsg.token import delay_mechanism as token

from vsg.vhdlFile import utils

from vsg.vhdlFile.classify import expression


def detect(iToken, lObjects):
    '''
    delay_mechanism ::=
        transport
      | [ reject *time*_expression ] inertial
    '''
    if utils.is_next_token_one_of(['transport', 'reject', 'inertial'], iToken, lObjects):
        return classify(iToken, lObjects)
    return iToken


def classify(iToken, lObjects):

    if utils.is_next_token('transport', iToken, lObjects):

        return utils.assign_next_token_required('transport', token.transport_keyword, iToken, lObjects)

    else:

        iCurrent = iToken

        if utils.is_next_token('reject', iCurrent, lObjects):
            iCurrent = utils.assign_next_token_required('reject', token.reject_keyword, iCurrent, lObjects)
            iCurrent = expression.classify_until(['inertial'], iCurrent, lObjects)

        iCurrent = utils.assign_next_token_required('inertial', token.inertial_keyword, iCurrent, lObjects)
        return iCurrent




from vsg.rules import remove_excessive_blank_lines_above_line_starting_with_token

from vsg.token import port_clause as token

lTokens = []
lTokens.append(token.close_parenthesis)

class rule_024(remove_excessive_blank_lines_above_line_starting_with_token):
    '''
    This rule checks for blank lines before the close parenthesis in port declarations.

    **Violation**

    .. code-block:: vhdl

       port (
         WR_EN    : std_logic;
         RD_EN    : std_logic;
         OVERFLOW : std_logic;
         DATA     : inout std_logic


       );

    **Fix**

    .. code-block:: vhdl

       port (
         WR_EN    : in    std_logic;
         RD_EN    : in    std_logic;
         OVERFLOW : out   std_logic;
         DATA     : inout std_logic
       );
    '''
    def __init__(self):
        remove_excessive_blank_lines_above_line_starting_with_token.__init__(self, 'port', '024', lTokens, iAllow=0)
        self.solution = 'Remove blank lines above ).'
        self.configuration_documentation_link = None




from vsg import parser


class component_keyword(parser.keyword):
    '''
    unique_id = instantiated_unit : component_keyword
    '''

    def __init__(self, sString):
        parser.keyword.__init__(self, sString)


class component_name(parser.name):
    '''
    unique_id = instantiated_unit : component_name
    '''

    def __init__(self, sString):
        parser.name.__init__(self, sString)


class entity_keyword(parser.keyword):
    '''
    unique_id = instantiated_unit : entity_keyword
    '''

    def __init__(self, sString):
        parser.keyword.__init__(self, sString)


class library_name(parser.name):
    '''
    unique_id = instantiated_unit : library_name
    '''

    def __init__(self, sString):
        parser.name.__init__(self, sString)


class dot(parser.dot):
    '''
    unique_id = instantiated_unit : dot
    '''

    def __init__(self, sString):
        parser.dot.__init__(self, sString)


class entity_name(parser.name):
    '''
    unique_id = instantiated_unit : entity_name
    '''

    def __init__(self, sString):
        parser.name.__init__(self, sString)


class open_parenthesis(parser.open_parenthesis):
    '''
    unique_id = instantiated_unit : open_parenthesis
    '''

    def __init__(self, sString='('):
        parser.open_parenthesis.__init__(self)


class architecture_identifier(parser.identifier):
    '''
    unique_id = instantiated_unit : architecture_identifier
    '''

    def __init__(self, sString):
        parser.identifier.__init__(self, sString)


class close_parenthesis(parser.close_parenthesis):
    '''
    unique_id = instantiated_unit : close_parenthesis
    '''

    def __init__(self, sString=')'):
        parser.close_parenthesis.__init__(self)


class configuration_keyword(parser.keyword):
    '''
    unique_id = instantiated_unit : configuration_keyword
    '''

    def __init__(self, sString):
        parser.keyword.__init__(self, sString)


class configuration_name(parser.name):
    '''
    unique_id = instantiated_unit : configuration_name
    '''

    def __init__(self, sString):
        parser.name.__init__(self, sString)





from vsg import parser


class record_element_simple_name(parser.name):
    '''
    unique_id = record_element_constraint : record_element_simple_name
    '''

    def __init__(self, sString):
        parser.name.__init__(self, sString)




from vsg.token import case_statement as token

from vsg.vhdlFile import utils

from vsg.vhdlFile.classify import case_statement_alternative
from vsg.vhdlFile.classify import expression


def detect(iToken, lObjects):
    '''
    case_statement ::=
        [ *case*_label : ]
            case [ ? ] expression is
                case_statement_alternative
                { case_statement_alternative }
        end case [ ? ] [ case_label ] ;
    '''
    if utils.keyword_found('case', iToken, lObjects):
        return classify(iToken, lObjects)
    return iToken


def classify(iToken, lObjects):

    iCurrent = utils.tokenize_label(iToken, lObjects, token.case_label, token.label_colon)
    iCurrent = utils.assign_next_token_required('case', token.case_keyword, iCurrent, lObjects)
    iCurrent = utils.assign_next_token_if('?', token.question_mark, iCurrent, lObjects)

    iCurrent = expression.classify_until(['is'], iCurrent, lObjects)

    iCurrent = utils.assign_next_token_required('is', token.is_keyword, iCurrent, lObjects)

    iCurrent = utils.detect_submodule(iCurrent, lObjects, case_statement_alternative)

    iCurrent = utils.assign_next_token_required('end', token.end_keyword, iToken, lObjects)
    iCurrent = utils.assign_next_token_required('case', token.end_case_keyword, iCurrent, lObjects)
    iCurrent = utils.assign_next_token_if('?', token.question_mark, iCurrent, lObjects)
    iCurrent = utils.assign_next_token_if_not(';', token.end_case_label, iCurrent, lObjects)
    iCurrent = utils.assign_next_token_required(';', token.semicolon, iCurrent, lObjects)

    return iCurrent




from vsg import parser

from vsg.token import pragma


def classify(lTokens, lObjects, lOpenPragmas, lClosePragmas, dVars):
    '''
    Classifies pragmas

    '''
    for iToken, sToken in enumerate(lTokens):
        if sToken in lOpenPragmas:
            dVars['pragma'] = True
        if dVars['pragma']:
            if not isinstance(lObjects[iToken], parser.whitespace):
                lObjects[iToken] = pragma.ignore(sToken)
        if sToken in lClosePragmas:
            dVars['pragma'] = False




from vsg.token import signature as token

from vsg.vhdlFile import utils

from vsg.vhdlFile.classify import type_mark


def detect(iToken, lObjects):
    '''
    signature ::= **[** [ type_mark { , type_mark } ] [ return type_mark ] **]**

    NOTE:  The [ and ] enclosed by ** are required if the signature is provided.
    '''

    if utils.is_next_token('[', iToken, lObjects):
        return classify(iToken, lObjects)

    return iToken


def classify(iToken, lObjects):

    iCurrent = utils.assign_next_token_required('[', token.open_bracket, iToken, lObjects)

    detect_type_mark(iCurrent, lObjects)

    detect_return(iCurrent, lObjects)

    iCurrent = utils.assign_next_token_required(']', token.close_bracket, iCurrent, lObjects)

    return iCurrent


def detect_return(iToken, lObjects):

    iCurrent = iToken
    if utils.is_next_token('return', iCurrent, lObjects):
        iCurrent = utils.assign_next_token_required('return', token.return_keyword, iCurrent, lObjects)
        iCurrent = type_mark.classify(iCurrent, lObjects)
    return iCurrent


def detect_type_mark(iToken, lObjects):

    iCurrent = iToken
    if not utils.is_next_token('return', iCurrent, lObjects):
        iCurrent = type_mark.classify(iCurrent, lObjects)
        while utils.is_next_token(',', iCurrent, lObjects):
            iCurrent = utils.assign_next_token_required(',', token.comma, iCurrent, lObjects)
            iCurrent = type_mark.classify(iCurrent, lObjects)

    return iCurrent




from vsg.rules import move_token_next_to_another_token

from vsg.token import port_clause as token


class rule_021(move_token_next_to_another_token):
    '''
    This rule checks the **port** keyword is on the same line as the (.

    **Violation**

    .. code-block:: vhdl

       port
       (

    **Fix**

    .. code-block:: vhdl

       port (
    '''

    def __init__(self):
        move_token_next_to_another_token.__init__(self, 'port', '021', token.port_keyword, token.open_parenthesis)
        self.solution = 'Move the ( to the same line as the *port* keyword.'




from vsg.rules import align_tokens_in_region_between_tokens

from vsg import token

lAlign = []
lAlign.append(token.concurrent_simple_signal_assignment.assignment)
lAlign.append(token.concurrent_conditional_signal_assignment.assignment)


class rule_006(align_tokens_in_region_between_tokens):
    '''
    This rule checks the alignment of the **<=** operator over multiple consecutive lines.
    |configuring_keyword_alignment_rules_link|

    **Violation**

    .. code-block:: vhdl

       wr_en <= '0';
       rd_en   <= '1';
       data <= (others => '0');

    **Fix**

    .. code-block:: vhdl

       wr_en <= '0';
       rd_en <= '1';
       data  <= (others => '0');
    '''

    def __init__(self):
        align_tokens_in_region_between_tokens.__init__(self, 'concurrent', '006', lAlign, token.architecture_body.begin_keyword, token.architecture_body.end_keyword)
        self.solution = 'Inconsistent alignment of "<=" in group of lines.'
        self.generate_statement_ends_group = True
        self.configuration.append('generate_statement_ends_group')




from vsg.rules.whitespace_between_token_pairs import Rule

from vsg.token import context_reference as token

lTokenPairs = []
lTokenPairs.append([token.keyword, token.library_name])
lTokenPairs.append([token.keyword, token.context_name])


class rule_002(Rule):
    '''
    This rule checks for a single space between the **context** keyword and the context selected name.

    |configuring_whitespace_rules_link|

    **Violation**

    .. code-block:: vhdl

       context   c1;

    **Fix**

    .. code-block:: vhdl

       context c1;
    '''
    def __init__(self):
        Rule.__init__(self, 'context_ref', '002', lTokenPairs)




from vsg.token import alias_declaration as token

from vsg.vhdlFile import utils

from vsg.vhdlFile.classify import name
from vsg.vhdlFile.classify import signature
from vsg.vhdlFile.classify import subtype_indication


def detect(iToken, lObjects):
    '''
    alias_declaration ::=
        alias alias_designator [ : subtype_indication ] is name [ signature ] ;
    '''

    if utils.is_next_token('alias', iToken, lObjects):
        return classify(iToken, lObjects)

    return iToken


def classify(iToken, lObjects):

    iCurrent = utils.assign_next_token_required('alias', token.alias_keyword, iToken, lObjects)

    iCurrent = utils.assign_next_token(token.alias_designator, iCurrent, lObjects)

    if utils.is_next_token(':', iCurrent, lObjects):
        iCurrent = utils.assign_next_token_required(':', token.colon, iCurrent, lObjects)
        iCurrent = subtype_indication.classify(iCurrent, lObjects)
    iCurrent = utils.assign_next_token_required('is', token.is_keyword, iCurrent, lObjects)

    iCurrent = name.classify_until([';', '['], iCurrent, lObjects)

    iCurrent = signature.detect(iCurrent, lObjects)

    iCurrent = utils.assign_next_token_required(';', token.semicolon, iCurrent, lObjects)

    return iCurrent




from vsg.rules import token_prefix_between_tokens

from vsg import token

lTokens = []
lTokens.append(token.interface_constant_declaration.identifier)
lTokens.append(token.interface_signal_declaration.identifier)
lTokens.append(token.interface_variable_declaration.identifier)
lTokens.append(token.interface_unknown_declaration.identifier)


class rule_011(token_prefix_between_tokens):
    '''
    This rule checks for valid prefixes on port identifiers.
    The default port prefixes are: *i\_*, *o\_*, *io\_*.

    |configuring_prefix_and_suffix_rules_link|

    **Violation**

    .. code-block:: vhdl

       port (
         wr_en    : in    std_logic;
         rd_en    : in    std_logic;
         overflow : out   std_logic;
         data     : inout std_logic
       );


    **Fix**

    .. code-block:: vhdl

       port (
         i_wr_en    : in    std_logic;
         i_rd_en    : in    std_logic;
         o_overflow : out   std_logic;
         io_data    : inout std_logic
       );
    '''

    def __init__(self):
        token_prefix_between_tokens.__init__(self, 'port', '011', lTokens, token.port_clause.open_parenthesis, token.port_clause.close_parenthesis)
        self.prefixes = ['i_', 'o_', 'io_']




from vsg.rules import align_tokens_in_region_between_tokens

from vsg import token

lAlign = []
lAlign.append(token.interface_unknown_declaration.assignment)


class rule_018(align_tokens_in_region_between_tokens):
    '''
    This rule checks the alignment of **:=** operator for each generic and port in the entity declaration.

    Following extra configurations are supported:

    * :code:`separate_generic_port_alignment`.

    |configuring_keyword_alignment_rules_link|

    **Violation**

    .. code-block:: vhdl

       generic (
           g_width        : positive := 8;
           g_output_delay : positive      := 5
       );
       port (
           clk_i   : in std_logic;
           data1_i : in std_logic  := 'X';
           data2_i : in std_logic      := 'X';
           data_o  : in std_logic
       );

    **Fix**

    .. code-block:: vhdl

       generic (
           g_width        : positive := 8;
           g_output_delay : positive := 5
       );
       port (
           clk_i   : in std_logic;
           data1_i : in std_logic := 'X';
           data2_i : in std_logic := 'X';
           data_o  : in std_logic
       );
    '''

    def __init__(self):
        align_tokens_in_region_between_tokens.__init__(self, 'entity', '018', lAlign, token.entity_declaration.entity_keyword, token.entity_declaration.end_keyword)
        self.solution = 'Align :.'




from vsg import parser
from vsg import token

from vsg.rules import consistent_token_case

lTokens = []
lTokens.append(token.subtype_declaration.identifier)

lIgnore = []
lIgnore.append(token.interface_signal_declaration.identifier)
lIgnore.append(token.interface_unknown_declaration.identifier)
lIgnore.append(token.interface_constant_declaration.identifier)
lIgnore.append(token.interface_variable_declaration.identifier)
lIgnore.append(token.association_element.formal_part)
lIgnore.append(parser.whitespace)
lIgnore.append(parser.carriage_return)
lIgnore.append(parser.blank_line)


class rule_002(consistent_token_case):
    '''
    This rule checks for consistent capitalization of subtype names.

    **Violation**

    .. code-block:: vhdl

       subtype read_size is range 0 to 9;
       subtype write_size is range 0 to 9;

       signal read  : READ_SIZE;
       signal write : write_size;

       constant read_sz  : read_size := 8;
       constant write_sz : WRITE_size := 1;


    **Fix**

    .. code-block:: vhdl

       subtype read_size is range 0 to 9;
       subtype write_size is range 0 to 9;

       signal read  : read_size;
       signal write : write_size;

       constant read_sz  : read_size := 8;
       constant write_sz : write_size := 1;
    '''

    def __init__(self):
        consistent_token_case.__init__(self, 'subtype', '002', lTokens, lIgnore)





from vsg import parser
from vsg import violation

from vsg.rules import utils as rules_utils
from vsg.rule_group import structure
from vsg.vhdlFile import utils


class insert_carriage_return_after_token_if_it_is_not_followed_by_a_comment_and_not_on_same_line_as_token(structure.Rule):
    '''
    Checks function parameters are on their own line except if they are all on the same line.

    Parameters
    ----------

    name : string
       The group the rule belongs to.

    identifier : string
       unique identifier.  Usually in the form of 00N.
    '''

    def __init__(self, name, identifier, token, oSameLineToken):
        structure.Rule.__init__(self, name=name, identifier=identifier)
        self.token = token
        self.oSameLineToken = oSameLineToken
        self.configuration_documentation_link = None

    def _get_tokens_of_interest(self, oFile):
        return oFile.get_tokens_bounded_by(self.token, parser.carriage_return, include_trailing_whitespace=True)

    def _analyze(self, lToi):
        for oToi in lToi:
           lTokens = oToi.get_tokens()
           if utils.are_next_consecutive_token_types([parser.carriage_return], 1, lTokens):
               continue
           if utils.are_next_consecutive_token_types([parser.whitespace, parser.comment], 1, lTokens):
               continue
           if utils.are_next_consecutive_token_types([parser.comment], 1, lTokens):
               continue
           for oToken in lTokens:
               if isinstance(oToken, self.oSameLineToken):
                   break
           else:
               self.add_violation(violation.New(oToi.get_line_number(), oToi, self.solution))

    def _fix_violation(self, oViolation):
        lTokens = oViolation.get_tokens()
        rules_utils.insert_carriage_return(lTokens, 1)
        oViolation.set_tokens(lTokens)




from vsg.rules import insert_token_left_of_token_if_it_does_not_exist_between_tokens_using_value_from_token

from vsg.token import process_statement as token

oInsertToken = token.end_process_label

oAnchorToken = token.semicolon

oLeftToken = token.end_keyword
oRightToken = token.semicolon

oValueToken = token.process_label


class rule_018(insert_token_left_of_token_if_it_does_not_exist_between_tokens_using_value_from_token):
    '''
    This rule checks the **end process** line has a label.
    The closing label will be added if the opening process label exists.

    |configuring_optional_items_link|

    **Violation**

    .. code-block:: vhdl

       end process;

    **Fix**

    .. code-block:: vhdl

       end process proc_a;
    '''

    def __init__(self):
        insert_token_left_of_token_if_it_does_not_exist_between_tokens_using_value_from_token.__init__(self, 'process', '018', oInsertToken, oAnchorToken, oLeftToken, oRightToken, oValueToken)
        self.solution = 'a label for the "end process".'
        self.subphase = 2




from vsg.rules import multiline_alignment_between_tokens

from vsg import token

lTokenPairs = []
lTokenPairs.append([token.if_statement.if_keyword, token.if_statement.then_keyword])
lTokenPairs.append([token.if_statement.elsif_keyword, token.if_statement.then_keyword])


class rule_009(multiline_alignment_between_tokens):
    '''
    This rule checks the alignment of multiline boolean expressions.

    |configuring_multiline_indent_rules_link|

    **Violation**

    .. code-block:: vhdl

       if (a = '0' and b = '1' and
             c = '0') then

    **Fix**

    .. code-block:: vhdl

       if (a = '0' and b = '1' and
           c = '0') then
    '''

    def __init__(self):
        multiline_alignment_between_tokens.__init__(self, 'if', '009', lTokenPairs)
        self.phase = 4
        self.subphase = 2




from vsg import parser


class logical_operator(parser.keyword):
    '''
    unique_id = logical_operator : logical_operator
    '''

    def __init__(self, sString):
        parser.keyword.__init__(self, sString)


class and_operator(logical_operator):
    '''
    unique_id = logical_operator : and_operator
    '''

    def __init__(self, sString):
        logical_operator.__init__(self, sString)


class or_operator(logical_operator):
    '''
    unique_id = logical_operator : or_operator
    '''

    def __init__(self, sString):
        logical_operator.__init__(self, sString)


class nand_operator(logical_operator):
    '''
    unique_id = logical_operator : nand_operator
    '''

    def __init__(self, sString):
        logical_operator.__init__(self, sString)


class nor_operator(logical_operator):
    '''
    unique_id = logical_operator : nor_operator
    '''

    def __init__(self, sString):
        logical_operator.__init__(self, sString)


class xor_operator(logical_operator):
    '''
    unique_id = logical_operator : xor_operator
    '''

    def __init__(self, sString):
        logical_operator.__init__(self, sString)


class xnor_operator(logical_operator):
    '''
    unique_id = logical_operator : xnor_operator
    '''

    def __init__(self, sString):
        logical_operator.__init__(self, sString)




from vsg.token import full_type_declaration as token

from vsg.vhdlFile import utils

from vsg.vhdlFile.classify import identifier
from vsg.vhdlFile.classify import type_definition


def detect(iToken, lObjects):
    '''
    full_type_declaration ::=
        type identifier is type_definition ;
    '''

    if utils.are_next_consecutive_tokens(['type', None, 'is'], iToken, lObjects):
        return classify(iToken, lObjects)
    return iToken


def classify(iToken, lObjects):

    iCurrent = utils.assign_next_token_required('type', token.type_keyword, iToken, lObjects)

    iCurrent = identifier.classify(iCurrent, lObjects, token.identifier)

    iCurrent = utils.assign_next_token_required('is', token.is_keyword, iCurrent, lObjects)

    iCurrent = type_definition.detect(iCurrent, lObjects)

    iCurrent = utils.assign_next_token_required(';', token.semicolon, iCurrent, lObjects)

    return iCurrent




from vsg import proposed_rule


class rule_006(proposed_rule.Rule):
    '''
    This rule checks the semicolon is on the same line as the context selected name.

    .. NOTE:: This rule has not been implemented yet.

    **Violation**

    .. code-block:: vhdl

       context c1
       ;

       context
       c1
       ;

    **Fix**

    .. code-block:: vhdl

       context c1;

       context
       c1;
    '''

    def __init__(self):
        proposed_rule.Rule.__init__(self, 'context_ref', '006')




from vsg.rules import align_consecutive_lines_after_line_starting_with_token_and_stopping_with_token as Rule

from vsg.token import case_statement_alternative as token


class rule_011(Rule):
    '''
    This rule checks the alignment of multiline **when** statements.

    **Violation**

    .. code-block:: vhdl

       case data is

         when 0 | 1 | 2 | 3
          4 | 5 | 7 =>

    **Fix**

    .. code-block:: vhdl

       case data is

         when 0 | 1 | 2 | 3
              4 | 5 | 7 =>
    '''

    def __init__(self):
        Rule.__init__(self, 'case', '011')
        self.lStartTokens = [token.when_keyword]
        self.lEndTokens = [token.assignment]
        self.solution = 'Align one space after *when* keyword'
        self.configuration_documentation_link = None




from vsg.rules import token_case

from vsg import token

lTokens = []
lTokens.append(token.entity_name_list.all_keyword)


class rule_501(token_case):
    '''
    This rule checks the **all** keyword has proper case.

    |configuring_uppercase_and_lowercase_rules_link|

    **Violation**

    .. code-block:: vhdl

       attribute coordinate of ALL : component is (0.0, 17.5);

    **Fix**

    .. code-block:: vhdl

       attribute coordinate of all : component is (0.0, 17.5);
    '''

    def __init__(self):
        token_case.__init__(self, 'entity_specification', '501', lTokens)
        self.groups.append('case::keyword')




from vsg.token import component_declaration as token

from vsg.vhdlFile import utils

from vsg.vhdlFile.classify import generic_clause
from vsg.vhdlFile.classify import port_clause



def detect(iToken, lObjects):
    '''
    component_declaration ::=
        component identifier [ is ]
            [ *local*_generic_clause ]
            [ *local*_port_clause ]
        end component [ *component*_simple_name ] ;
    '''

    if utils.is_next_token('component', iToken, lObjects):
        return classify(iToken, lObjects)
    else:
        return iToken


def classify(iToken, lObjects):

    iCurrent = classify_opening_declaration(iToken, lObjects)

    iCurrent = generic_clause.detect(iCurrent, lObjects)

    iCurrent = port_clause.detect(iCurrent, lObjects)

    iCurrent = classify_closing_declaration(iCurrent, lObjects)

    return iCurrent


def classify_opening_declaration(iToken, lObjects):

    iCurrent = utils.assign_next_token_required('component', token.component_keyword, iToken, lObjects)
    iCurrent = utils.assign_next_token(token.identifier, iCurrent, lObjects)
    iCurrent = utils.assign_next_token_if('is', token.is_keyword, iCurrent, lObjects)

    return iCurrent


def classify_closing_declaration(iToken, lObjects):

    iCurrent = utils.assign_next_token_required('end', token.end_keyword, iToken, lObjects)
    iCurrent = utils.assign_next_token_required('component', token.end_component_keyword, iCurrent, lObjects)
    iCurrent = utils.assign_next_token_if_not(';', token.component_simple_name, iCurrent, lObjects)
    iCurrent = utils.assign_next_token_required(';', token.semicolon, iCurrent, lObjects)

    return iCurrent




from vsg.token import condition_clause as token

from vsg.vhdlFile import utils

from vsg.vhdlFile.classify import condition


def detect(iToken, lObjects):
    '''
    condition_clause ::=
        until condition
    '''

    if utils.is_next_token('until', iToken, lObjects):
        return True
    return False


def classify_until(lUntils, iToken, lObjects):

    iCurrent = utils.assign_next_token_required('until', token.until_keyword, iToken, lObjects)

    iCurrent = condition.classify_until(lUntils, iCurrent, lObjects)

    return iCurrent





from vsg.vhdlFile.classify import enumeration_type_definition
from vsg.vhdlFile.classify import physical_type_definition
from vsg.vhdlFile.classify import integer_type_definition


def detect(iToken, lObjects):
    '''
    scalar_type_definition ::=
        enumeration_type_definition
      | integer_type_definition
      | floating_type_definition
      | physical_type_definition

    NOTE:  floating and physical types are not parsed yet.
           They are very similar to integer types, and will hopefully not be required.
    '''

    iReturn = physical_type_definition.detect(iToken, lObjects)
    if iReturn != iToken:
        return iReturn

    iReturn = enumeration_type_definition.detect(iToken, lObjects)
    if iReturn != iToken:
        return iReturn

    iReturn = integer_type_definition.detect(iToken, lObjects)
    if iReturn != iToken:
        return iReturn

    return iToken




from vsg import parser
from vsg import violation

from vsg.rules import utils as rules_utils
from vsg.rule_group import naming


class token_prefix(naming.Rule):
    '''
    Checks the case for words.

    Parameters
    ----------

    name : string
       The group the rule belongs to.

    identifier : string
       unique identifier.  Usually in the form of 00N.

    lTokens : list of parser object types
       object types to check the prefix

    lPrefixes : string list
       acceptable prefixes
    '''

    def __init__(self, name, identifier, lTokens):
        naming.Rule.__init__(self, name=name, identifier=identifier)
        self.lTokens = lTokens
        self.prefixes = None
        self.configuration.append('prefixes')
        self.fixable = False
        self.disable = True

    def _get_tokens_of_interest(self, oFile):
        return oFile.get_tokens_matching(self.lTokens)

    def _analyze(self, lToi):
        lPrefixLower = []
        for sPrefix in self.prefixes:
            lPrefixLower.append(sPrefix.lower())
        for oToi in lToi:
            lTokens = oToi.get_tokens()
            sToken = lTokens[0].get_value().lower()
            bValid = False
            for sPrefix in lPrefixLower:
                if sToken.startswith(sPrefix.lower()):
                    bValid = True
            if not bValid:
                sSolution = 'Prefix ' + lTokens[0].get_value() + ' with one of the following: ' + ', '.join(self.prefixes)
                oViolation = violation.New(oToi.get_line_number(), oToi, sSolution)
                self.add_violation(oViolation)

    def _fix_violation(self, oViolation):
        lTokens = oViolation.get_tokens()
        if oViolation.get_action() == 'remove_whitespace':
            oViolation.set_tokens([lTokens[1]])
        elif oViolation.get_action() == 'adjust_whitespace':
            lTokens[0].set_value(lTokens[1].get_indent() * self.indentSize * ' ')
            oViolation.set_tokens(lTokens)
        elif oViolation.get_action() == 'add_whitespace':
            rules_utils.insert_whitespace(lTokens, 0, lTokens[0].get_indent() * self.indentSize)
            oViolation.set_tokens(lTokens)




from vsg import parser

from vsg.rules.whitespace_before_token import Rule

lTokens = []
lTokens.append(parser.comment)


class rule_004(Rule):
    '''
    This rule checks for at least a single space before inline comments.

    |configuring_whitespace_rules_link|

    **Violation**

    .. code-block:: vhdl

       wr_en <= '1';--Write data
       rd_en <= '1';   -- Read data

    **Fix**

    .. code-block:: vhdl

       wr_en <= '1'; --Write data
       rd_en <= '1';   -- Read data
    '''

    def __init__(self):
        Rule.__init__(self, 'comment', '004', lTokens)
        self.number_of_spaces = '>=1'




from vsg import parser

from vsg.vhdlFile import utils

from vsg.token import direction


def classify_until(lUntils, iToken, lObjects, oType=parser.todo):
    '''
      name ::=
              simple_name
            | operator_symbol
            | character_literal
            | selected_name
            | indexed_name
            | slice_name
            | attribute_name
            | external_name

    NOTE: At the moment, everything will be set to parser.todo.
    '''

    iCurrent = iToken
    iStop = len(lObjects) - 1
    iOpenParenthesis = 0
    iCloseParenthesis = 0
    while iCurrent < iStop:
        iCurrent = utils.find_next_token(iCurrent, lObjects)
        if utils.token_is_open_parenthesis(iCurrent, lObjects):
           iOpenParenthesis += 1
        if utils.token_is_close_parenthesis(iCurrent, lObjects):
           iCloseParenthesis += 1
        if iOpenParenthesis < iCloseParenthesis:
            break
        elif lObjects[iCurrent].get_value().lower() in lUntils:
            break
        else:
            utils.assign_special_tokens(lObjects, iCurrent, oType)
    return iCurrent




from vsg.vhdlFile.classify import interface_constant_declaration
from vsg.vhdlFile.classify import interface_file_declaration
from vsg.vhdlFile.classify import interface_signal_declaration
from vsg.vhdlFile.classify import interface_unknown_declaration
from vsg.vhdlFile.classify import interface_variable_declaration


def detect(iCurrent, lObjects):
    '''
    interface_object_declaration ::=
        interface_constant_declaration
      | interface_signal_declaration
      | interface_variable_declaration
      | interface_file_declaration
    '''

    iReturn = interface_constant_declaration.detect(iCurrent, lObjects)
    if iReturn != iCurrent:
        return iReturn

    iReturn = interface_signal_declaration.detect(iCurrent, lObjects)
    if iReturn != iCurrent:
        return iReturn

    iReturn = interface_variable_declaration.detect(iCurrent, lObjects)
    if iReturn != iCurrent:
        return iReturn

    iReturn = interface_file_declaration.detect(iCurrent, lObjects)
    if iReturn != iCurrent:
        return iReturn

    ### This captures constant, signal and variable declarations without optional keywords
    ### This is typically done in port lists
    iReturn = interface_unknown_declaration.detect(iCurrent, lObjects)
    if iReturn != iCurrent:
        return iReturn

    return iCurrent




from vsg.rules import remove_excessive_blank_lines_above_line_starting_with_token

from vsg.token import generic_clause as token

lTokens = []
lTokens.append(token.close_parenthesis)


class rule_019(remove_excessive_blank_lines_above_line_starting_with_token):
    '''
    This rule checks for blank lines before the ); of the generic declaration.

    **Violation**

    .. code-block:: vhdl

      generic (
        g_width : std_logic := '0';
        g_depth : Std_logic := '1'


      );

    **Fix**

    .. code-block:: vhdl

      generic (
        g_width : std_logic := '0';
        g_depth : Std_logic := '1'
      );
    '''
    def __init__(self):
        remove_excessive_blank_lines_above_line_starting_with_token.__init__(self, 'generic', '019', lTokens, iAllow=0)
        self.configuration_documentation_link = None




from vsg.token import interface_list as token

from vsg.vhdlFile import utils

from vsg.vhdlFile.classify import interface_element


def classify(iToken, lObjects):
    '''
    interface_list ::=
        interface_element { ; interface_element }
    '''

    iCurrent = interface_element.classify(iToken, lObjects)

    while utils.is_next_token(';', iCurrent, lObjects):
         iCurrent = utils.assign_next_token_required(';', token.semicolon, iCurrent, lObjects)
         iCurrent = interface_element.classify(iToken, lObjects)

    return iCurrent




from vsg.rules import blank_line_above_line_starting_with_token

from vsg import token

lTokens = []
lTokens.append(token.subprogram_body.begin_keyword)

lAllowTokens = []
lAllowTokens.append(token.subprogram_body.is_keyword)


class rule_202(blank_line_above_line_starting_with_token):
    '''
    This rule checks for blank lines above the **begin** keyword.

    This rule allows the **is** keyword to occupy the blank line:

    .. code-block:: vhdl

       function overflow (a: integer) return integer is
       begin

    |configuring_blank_lines_link|

    **Violation**

    .. code-block:: vhdl

       function overflow (a: integer) return integer is

         constant width : integer := 32;
       begin

    **Fix**

    .. code-block:: vhdl

       function overflow (a: integer) return integer is

         constant width : integer := 32;

       begin
    '''

    def __init__(self):
        blank_line_above_line_starting_with_token.__init__(self, 'subprogram_body', '202', lTokens, lAllowTokens)




from vsg.token import physical_type_definition as token

from vsg.vhdlFile import utils

from vsg.vhdlFile.classify import primary_unit_declaration
from vsg.vhdlFile.classify import range_constraint
from vsg.vhdlFile.classify import secondary_unit_declaration


def detect(iToken, lObjects):
    '''
    physical_type_definition ::=
        range_constraint
            **units**
                primary_unit_declaration
                { secondary_unit_declaration }
            **end** **units** [ physical_type_simple_name ]
    '''
    if units_keyword_found_before_semicolon(iToken, lObjects):
        return classify(iToken, lObjects)
    return iToken


def classify(iToken, lObjects):
    iCurrent = range_constraint.detect(iToken, lObjects)

    iCurrent = utils.assign_next_token_required('units', token.units_keyword, iToken, lObjects)

    iCurrent = primary_unit_declaration.detect(iCurrent, lObjects)

    while not utils.is_next_token('end', iCurrent, lObjects):
        iCurrent = secondary_unit_declaration.detect(iCurrent, lObjects)

    iCurrent = utils.assign_next_token(token.end_keyword, iCurrent, lObjects)
    iCurrent = utils.assign_next_token_required('units', token.end_units_keyword, iCurrent, lObjects)
    iCurrent = utils.assign_next_token_if_not(';', token.simple_name, iCurrent, lObjects)

    return iCurrent


def units_keyword_found_before_semicolon(iToken, lObjects):
    if utils.find_in_range('units', iToken, ';', lObjects):
        return True
    return False




from vsg import token

from vsg.rules.whitespace_between_token_pairs import Rule

lTokens = []
lTokens.append([token.if_statement.end_keyword, token.if_statement.end_if_keyword])


class rule_015(Rule):
    '''
    This rule checks for a single space between the **end if** keywords.

    |configuring_whitespace_rules_link|

    **Violation**

    .. code-block:: vhdl

       end    if;

    **Fix**

    .. code-block:: vhdl

       end if;
    '''
    def __init__(self):
        Rule.__init__(self, 'if', '015', lTokens)




from vsg.rules import insert_carriage_return_after_token_if_it_is_not_followed_by_a_comment as Rule

from vsg.token import entity_declaration as token

lTokens = []
lTokens.append(token.begin_keyword)


class rule_027(Rule):
    '''
    This rule checks for code after the **begin** keyword.

    **Violation**

    .. code-block:: vhdl

       begin end entity;

    **Fix**

    .. code-block:: vhdl

        begin
        end entity;
    '''

    def __init__(self):
        Rule.__init__(self, 'entity', '027', lTokens)
        self.solution = 'Move code after the **begin** keyword to the next line.'




from vsg import parser
from vsg import token

from vsg.vhdlFile import utils


def remove_optional_item(oViolation, oInsertToken):
    lTokens = oViolation.get_tokens()
    if isinstance(lTokens[0], parser.whitespace):
        oViolation.set_tokens([])
    else:
        oViolation.set_tokens([lTokens[0]])


def is_single_line(oToi):
    lTokens = oToi.get_tokens()
    for oToken in lTokens:
        if isinstance(oToken, parser.carriage_return):
            return False
    return True


def number_of_carriage_returns(lTokens):
    iReturn = 0
    for oToken in lTokens:
        if isinstance(oToken, parser.carriage_return):
            iReturn += 1
    return iReturn


def number_of_tokens_from_token_list_in_token_list(lTokens, lTokenList):
    iReturn = 0
    for oToken in lTokens:
        iReturn += number_of_tokens_in_token_list(oToken, lTokenList)
    return iReturn


def number_of_tokens_in_token_list(oToken, lTokens):
    iReturn = 0
    for oTokenItem in lTokens:
        if oToken == type(oTokenItem):
            iReturn += 1
    return iReturn


def does_line_start_with_comment(lTokens):
    if isinstance(lTokens[0], parser.comment):
        return True
    if isinstance(lTokens[0], parser.whitespace) and isinstance(lTokens[1], parser.comment):
        return True
    return False


def get_toi_parameters(oToi):
    return oToi.get_line_number(), oToi.get_tokens()


def insert_token(lTokens, index, oToken):
    try:
        oToken = update_code_tags(lTokens[index], oToken)
    except IndexError:
        oToken = update_code_tags(lTokens[0], oToken)

    lTokens.insert(index, oToken)


def append_token(lTokens, oToken):
    oToken = update_code_tags(lTokens[-1], oToken)
    lTokens.append(oToken)


def update_code_tags(oToken1, oToken2):
    oToken2.code_tags = oToken1.code_tags
    return oToken2


def insert_whitespace(lTokens, index, num=1, sString=' '):
    if sString == ' ':
        insert_token(lTokens, index, parser.whitespace(' '*num))
    else:
        oToken = parser.whitespace('\t'*num)
        oToken.has_tabs = True
        insert_token(lTokens, index, oToken)


def insert_new_whitespace(lTokens, index, sWhitespace):
    oToken = parser.whitespace(sWhitespace)
    if '\t' in sWhitespace:
        oToken.has_tabs = True
    insert_token(lTokens, index, oToken)


def insert_carriage_return(lTokens, index):
    insert_token(lTokens, index, parser.carriage_return())


def insert_blank_line(lTokens, index):
    insert_token(lTokens, index, parser.blank_line())


def append_whitespace(lTokens, num=1):
    append_token(lTokens, parser.whitespace(' '*num))


def append_carriage_return(lTokens):
    append_token(lTokens, parser.carriage_return())


def get_index_of_token_in_list(oToken, lTokens):
    for iToken, token in enumerate(lTokens):
        if isinstance(token, oToken):
            return iToken
    return None


def get_indexes_of_token_in_list(oToken, lTokens):
    lReturn = []
    for iToken, token in enumerate(lTokens):
        if isinstance(token, oToken):
            lReturn.append(iToken)
    return lReturn


def get_number_of_carriage_returns_before_token(oStopToken, lTokens):
    iReturn = 0
    for oToken in lTokens:
        if isinstance(oToken, parser.carriage_return):
            iReturn += 1
        if isinstance(oToken, oStopToken):
            break
    return iReturn


def whitespace_before_token_index(lTokens, iIndex):
    if isinstance(lTokens[iIndex - 1], parser.whitespace):
        return True
    return False


def remove_token_sequence_from_token_list(lRemoveTokens, lTokens):
    lSliceIndexes = find_slice_indexes_of_token_sequence_in_token_list(lRemoveTokens, lTokens)
    return remove_slices_from_token_list(lTokens, lSliceIndexes)


def find_slice_indexes_of_token_sequence_in_token_list(lRemoveTokens, lTokens):
    iTokenListLength = len(lTokens)
    iLength = len(lRemoveTokens)
    lReturn = []
    for iIndex in range(0, iTokenListLength - 1 - iLength):
        if is_token_sequence_at_index_in_token_list(lRemoveTokens, iIndex, lTokens):
            lReturn.append([iIndex, iIndex + iLength])
    return lReturn


def is_token_sequence_at_index_in_token_list(lRemoveTokens, iIndex, lTokens):
    iLength = len(lRemoveTokens)
    for iToken in range(0, iLength):
        if not isinstance(lTokens[iIndex + iToken], lRemoveTokens[iToken]):
            return False
    return True


def remove_slices_from_token_list(lTokens, lSlices):
    lReturn = lTokens
    lSlices.reverse()
    for lSlice in lSlices:
        lReturn = remove_slice_from_token_list(lReturn, lSlice)
    return lReturn


def remove_slice_from_token_list(lTokens, lSlice):
    iStart = lSlice[0]
    iEnd = lSlice[1]
    return lTokens[0:iStart] + lTokens[iEnd::]


def token_is_carriage_return(oToken):
    if isinstance(oToken, parser.carriage_return):
        return True
    return False


def token_is_comment(oToken):
    if isinstance(oToken, parser.comment):
        return True
    return False


def token_is_blank_line(oToken):
    if isinstance(oToken, parser.blank_line):
        return True
    return False


def token_is_open_paren(oToken):
    if isinstance(oToken, parser.open_parenthesis):
        return True
    return False


def token_is_close_paren(oToken):
    if isinstance(oToken, parser.close_parenthesis):
        return True
    return False


def token_is_parenthesis(oToken):
    if token_is_open_paren(oToken) or token_is_close_paren(oToken):
        return True
    return False


def token_is_whitespace(oToken):
    if isinstance(oToken, parser.whitespace):
        return True
    return False


def token_list_begins_with_close_paren(lTokens):
    if isinstance(lTokens[0], parser.whitespace) and isinstance(lTokens[1], parser.close_parenthesis):
        return True
    if isinstance(lTokens[0], parser.close_parenthesis):
        return True
    return False


def token_list_starts_with_paren(lTokens, iIndex=0):
    iToken = utils.find_next_non_whitespace_token(iIndex, lTokens)
    if isinstance(lTokens[iToken], parser.open_parenthesis):
        return True
    return False


def token_at_the_beginning_of_a_line(oToken, lTokens):
    if isinstance(lTokens[0], parser.whitespace) and isinstance(lTokens[1], oToken):
        return True
    if isinstance(lTokens[0], oToken):
        return True
    return False


def token_at_beginning_of_line_in_token_list(iToken, lTokens):
    if isinstance(lTokens[iToken - 1], parser.carriage_return):
        return True
    if isinstance(lTokens[iToken - 1], parser.whitespace) and isinstance(lTokens[iToken - 2], parser.carriage_return):
        return True
    return False


def token_list_is_the_beginning_of_a_line(lTokens):
    if isinstance(lTokens[0], parser.carriage_return) and isinstance(lTokens[1], parser.whitespace):
        return True
    if isinstance(lTokens[1], parser.carriage_return):
        return True
    return False


def left_most_token_is_at_the_end_of_a_line(lTokens):
    if token_is_carriage_return(lTokens[1]):
        return True
    if token_is_comment(lTokens[1]):
        return True
    if token_is_whitespace(lTokens[1]) and token_is_comment(lTokens[2]):
        return True
    return False


def whitespace_is_larger_than_a_single_character(lTokens):
    if lTokens[1].get_value() != ' ':
        return True
    return False


def remove_toi_if_token_is_at_the_end_of_the_line(lToi):
    lReturn = []
    for oToi in lToi:
        lTokens = oToi.get_tokens()
        if left_most_token_is_at_the_end_of_a_line(lTokens):
            continue
        lReturn.append(oToi)
    return lReturn


def lowercase_list(lList):
    lReturn = []
    for sItem in lList:
        lReturn.append(sItem.lower())
    return lReturn
  

def extract_identifiers_with_mode_of_input(lToi):
    return extract_identifiers_with_mode(lToi, token.mode.in_keyword)


def extract_identifiers_with_mode_of_out(lToi):
    return extract_identifiers_with_mode(lToi, token.mode.out_keyword)


def extract_identifiers_with_mode_of_inout(lToi):
    return extract_identifiers_with_mode(lToi, token.mode.inout_keyword)


def extract_identifiers_with_mode_of_buffer(lToi):
    return extract_identifiers_with_mode(lToi, token.mode.buffer_keyword)


def extract_identifiers_with_mode_of_linkage(lToi):
    return extract_identifiers_with_mode(lToi, token.mode.linkage_keyword)


def extract_identifiers_with_mode(lToi, oTokenType):
    lReturn = []
    for oToi in lToi:
        if oToi.token_type_exists(oTokenType):
            iIndex = oToi.get_index_of_token_matching(parser.identifier)
            lReturn.append(oToi.extract_tokens(iIndex, iIndex))
    return lReturn


def remove_leading_whitespace_tokens(lTokens):
    if len(lTokens) > 1 and isinstance(lTokens[0], parser.whitespace):
        lTokens.pop(0)


def change_all_whitespace_to_single_character(lTokens):
    for oToken in lTokens:
        if isinstance(oToken, parser.whitespace):
            oToken.set_value(' ')


def token_is_at_beginning_of_line(lTokens):
    if isinstance(lTokens[0], parser.carriage_return):
        return True
    if isinstance(lTokens[1], parser.carriage_return):
        return True
    return False


def update_open_paren_counter(oToken, iOpenParen):
    if token_is_open_paren(oToken):
        return iOpenParen + 1
    return iOpenParen


def update_close_paren_counter(oToken, iCloseParen):
    if token_is_close_paren(oToken):
        return iCloseParen + 1
    return iCloseParen


def update_paren_counter(oToken, iParen):
    if token_is_open_paren(oToken):
        iParen += 1
    elif token_is_close_paren(oToken):
        iParen -= 1
    return iParen


def analyze_with_function(self, oToi, oTokenType, fFunction):
    iLine, lTokens = get_toi_parameters(oToi)

    for iToken, oToken in enumerate(lTokens):
        iLine = utils.increment_line_number(iLine, oToken)
        if isinstance(oToken, oTokenType):
            oToi.set_meta_data('iStartLine', iLine)
            oToi.set_meta_data('iStart', iToken)
            oToi.set_meta_data('iToken', iToken)
            fFunction(self, oToi)

def token_exists_in_token_type_list(oToken, lTypeTokens):
    for oTokenType in lTypeTokens:
        if isinstance(oToken, oTokenType):
            return True
    return False

def is_next_token_ignoring_whitespace(oToken, iToken, lTokens):
    iToken = utils.find_next_non_whitespace_token(iToken + 1, lTokens)
    if isinstance(lTokens[iToken], oToken):
        return True
    return False


def array_detected_after_assignment_operator(assignment_operator, oToi):
    lTokens = oToi.get_tokens()
    if not open_paren_after_assignment_operator(assignment_operator, lTokens):
        return False
    if open_paren_after_assignment_operator_is_aggregate(assignment_operator, lTokens):
        return True
    if open_paren_after_assignment_operator_id_matches_last_paren_id(assignment_operator, lTokens):
        return True
    if open_paren_after_assignment_operator_is_followed_by_aggregate_open_parenthesis(assignment_operator, lTokens):
        return True
    if not close_paren_detected_at_end_of_tokens(lTokens):
        return False

    iParen = 0
    bFirstTokenFound = False
    iParenId = -1
    for oToken in lTokens:
        if iParenId == -1 and token_is_open_paren(oToken):
            iParenId = oToken.iId
        if second_open_paren_detected_after_first_open_paren_was_closed(bFirstTokenFound, iParen, oToken):
            return False
        iParen = update_paren_counter(oToken, iParen)
        if not bFirstTokenFound:
            bFirstTokenFound = token_is_open_paren(oToken)
#        print(f'{iParen}|{bFirstTokenFound}|{oToken}')
    
    return True


def second_open_paren_detected_after_first_open_paren_was_closed(bFirstTokenFound, iParen, oToken):
    if bFirstTokenFound:
        if iParen == 0 and token_is_open_paren(oToken):
            return True
    return False


def open_paren_after_assignment_operator(assignment_operator, lTokens):
    iToken = get_index_of_token_in_list(assignment_operator, lTokens)
    return is_next_token_ignoring_whitespace(parser.open_parenthesis, iToken, lTokens)


def open_paren_after_assignment_operator_is_aggregate(assignment_operator, lTokens):
    iToken = get_index_of_token_in_list(assignment_operator, lTokens)
    if is_next_token_ignoring_whitespace(parser.open_parenthesis, iToken, lTokens):
        iIndex = get_index_of_token_in_list(parser.open_parenthesis, lTokens)
        return isinstance(lTokens[iIndex], token.aggregate.open_parenthesis)
    return False


def open_paren_after_assignment_operator_id_matches_last_paren_id(assignment_operator, lTokens):
    iToken = get_index_of_token_in_list(assignment_operator, lTokens)
    if is_next_token_ignoring_whitespace(parser.open_parenthesis, iToken, lTokens):
        iIndex = get_index_of_token_in_list(parser.open_parenthesis, lTokens)
        iOpenParenId = lTokens[iIndex].iId
    lTokens.reverse()
    if is_next_token_ignoring_whitespace(parser.close_parenthesis, 1, lTokens):
        iIndex = get_index_of_token_in_list(parser.close_parenthesis, lTokens)
        iCloseParenId = lTokens[iIndex].iId
    else:
        lTokens.reverse()
        return False
    lTokens.reverse()
    if iOpenParenId == iCloseParenId:
        return True
    return False


def open_paren_after_assignment_operator_is_followed_by_aggregate_open_parenthesis(assignment_operator, lTokens):
    iToken = get_index_of_token_in_list(assignment_operator, lTokens)
    if is_next_token_ignoring_whitespace(parser.open_parenthesis, iToken, lTokens):
        iIndex = get_index_of_token_in_list(parser.open_parenthesis, lTokens)
        return is_next_token_ignoring_whitespace(token.aggregate.open_parenthesis, iIndex + 1, lTokens)
    return False


def close_paren_detected_at_end_of_tokens(lTokens):
    lTokens.reverse()
    if is_next_token_ignoring_whitespace(parser.close_parenthesis, 1, lTokens):
        lTokens.reverse()
        return True
    lTokens.reverse()
    return False




from vsg.rules.whitespace_between_tokens import Rule

from vsg.token import element_association
from vsg.token import choice


class rule_100(Rule):
    '''
    This rule checks for a single space between the **others** keyword and the => in an element_association.

    |configuring_whitespace_rules_link|

    **Violation**

    .. code-block:: vhdl

       a <= (others=> (others    => '0'));

    **Fix**

    .. code-block:: vhdl

       a <= (others => (others => '0'));
    '''
    def __init__(self):
        Rule.__init__(self, 'element_association', '100')
        self.left_token = choice.others_keyword
        self.right_token = element_association.assignment




from vsg.rules import move_token_next_to_another_token as Rule

from vsg.token import record_type_definition as token


class rule_004(Rule):
    '''
    This rule checks the **is** keyword is on the same line as the **record** keyword.

    **Violation**

    .. code-block:: vhdl

       type t_record is record
         a : std_logic;
         b : std_logic;
       end
       record;

    **Fix**

    .. code-block:: vhdl

       type t_record is record
         a : std_logic;
         b : std_logic;
       end record
       ;
    '''

    def __init__(self):
        Rule.__init__(self, 'record_type_definition', '004', token.end_keyword, token.end_record_keyword)
        self.subphase = 1
        self.solution = 'Move *record* next to *end* keyword'




from vsg.rules import align_tokens_in_region_between_tokens as Rule

from vsg import token

lAlign = []
lAlign.append(token.association_element.assignment)

oBegin = token.procedure_call.open_parenthesis
oEnd = token.procedure_call.close_parenthesis


class rule_401(Rule):
    '''
    This rule checks the alignment of :code:`=>` keywords in procedure calls.

    |configuring_keyword_alignment_rules_link|

    **Violation**

    .. code-block:: vhdl

       connect_ports(
         port_1=> data,
         port_2 => enable,
         port_3    => overflow,
         port_4  => underflow
       );

    **Fix**

    .. code-block:: vhdl

       connect_ports(
         port_1 => data,
         port_2 => enable,
         port_3 => overflow,
         port_4 => underflow
       );
    '''

    def __init__(self):
        Rule.__init__(self, 'procedure_call', '401', lAlign, oBegin, oEnd)
        self.solution = 'Align =>.'
        self.subphase = 2
        self.bIncludeTillBeginningOfLine = True




from vsg import token

from vsg.rules import move_token_left_to_next_non_whitespace_token as Rule

lTokens = []
lTokens.append(token.component_instantiation_statement.semicolon)


class rule_035(Rule):
    '''
    This rule checks the semicolon is not on its own line.

    **Violation**

    .. code-block:: vhdl

       U_FIFO : FIFO
         port map (
           A => B,
           B => C)
         ;

    **Fix**

    .. code-block:: vhdl

       U_FIFO : FIFO
         port map (
           A => B,
           B => C);
    '''

    def __init__(self):
        Rule.__init__(self, 'instantiation', '035', lTokens)
        self.bInsertWhitespace = False




from vsg.rules import blank_line_above_line_starting_with_token

from vsg.token import architecture_body as token


class rule_016(blank_line_above_line_starting_with_token):
    '''
    This rule checks for blank lines above the **begin** keyword.

    |configuring_blank_lines_link|

    **Violation**

    .. code-block:: vhdl

       architecture rtl of fifo is

         signal wr_en : std_logic;
       begin

    **Fix**

    .. code-block:: vhdl

       architecture rtl of fifo is

         signal wr_en : std_logic;

       begin
    '''

    def __init__(self):
        blank_line_above_line_starting_with_token.__init__(self, 'architecture', '016', [token.begin_keyword])




from vsg.rules import split_line_at_token

from vsg import token

lTokens = []
lTokens.append(token.case_generate_statement.end_keyword)
lTokens.append(token.for_generate_statement.end_keyword)
lTokens.append(token.if_generate_statement.end_keyword)


class rule_019(split_line_at_token):
    '''
    This rule checks the **end** keyword is on its own line.

    **Violation**

    .. code-block:: vhdl

       ram_array : for i in 0 to 7 generate
        a <= b; end generate;

    **Fix**

    .. code-block:: vhdl

       ram_array : for i in 0 to 7 generate
         a <= b;
       end generate;
    '''

    def __init__(self):
        split_line_at_token.__init__(self, 'generate', '019', lTokens)
        self.solution = 'Move end keyword and any code after it to the next line'




from vsg.vhdlFile.classify import expression


def classify_until(lUntils, iToken, lObjects):
    '''
    condition ::=
        expression
    '''
    return expression.classify_until(lUntils, iToken, lObjects)




from vsg.rules import token_suffix as Rule

from vsg.rules import utils

from vsg import token

lTokens = []


class rule_608(Rule):
    '''
    This rule checks for valid suffixes on port identifiers for buffer ports.

    The default suffix is: *_b*.

    |configuring_prefix_and_suffix_rules_link|

    **Violation**

    .. code-block:: vhdl

       port (
         wr_en    : buffer    std_logic;
         rd_en    : buffer    std_logic
       );

    **Fix**

    .. code-block:: vhdl

       port (
         wr_en_b    : buffer    std_logic;
         rd_en_b    : buffer    std_logic
       );
    '''

    def __init__(self):
        Rule.__init__(self, 'port', '608', lTokens)
        self.suffixes = ['_b']

    def _get_tokens_of_interest(self, oFile):
        lReturn = []
        lToi = oFile.get_interface_elements_between_tokens(token.port_clause.open_parenthesis, token.port_clause.close_parenthesis)
        lReturn = utils.extract_identifiers_with_mode_of_buffer(lToi)
        return lReturn




from vsg import violation

from vsg.vhdlFile import utils
from vsg.rule_group import case


class consistent_token_case(case.Rule):
    '''
    Checks the case for words.

    Parameters
    ----------

    name : string
       The group the rule belongs to.

    identifier : string
       unique identifier.  Usually in the form of 00N.

    lTokens: list of token type objects
       token type to apply the case check against
    '''

    def __init__(self, name, identifier, lTokens, lIgnore=None):
        case.Rule.__init__(self, name=name, identifier=identifier)
        self.subphase = 2
        self.lTokens = lTokens
        if lIgnore == None:
            self.lIgnoreTokens = []
        else:
            self.lIgnoreTokens = lIgnore
        self.configuration_documentation_link = None

    def analyze(self, oFile):
        lTargetTypes = oFile.get_tokens_matching(self.lTokens)
        lTargetValues = []
        lTargetValuesLower = []
        for oTargetType in lTargetTypes:
            oToken = oTargetType.get_tokens()[0]
            lTargetValues.append(oToken.get_value())
            lTargetValuesLower.append(oToken.get_value().lower())

        oToi = oFile.get_all_tokens()
        iLine, lTokens = utils.get_toi_parameters(oToi)

        for iToken, oToken in enumerate(lTokens):

            iLine = utils.increment_line_number(iLine, oToken)

            if is_token_in_ignore_token_list(oToken, self.lIgnoreTokens):
                continue

            sTokenValue = oToken.get_value()
            sTokenValueLower = sTokenValue.lower()
            for sTargetValue, sTargetValueLower in zip(lTargetValues, lTargetValuesLower):
                if sTokenValueLower == sTargetValueLower:
                    if sTokenValue != sTargetValue:
                        sSolution = 'Change "' + sTokenValue + '" to "' + sTargetValue + '"'
                        oNewToi = oToi.extract_tokens(iToken, iToken)
                        oViolation = violation.New(iLine, oNewToi, sSolution)
                        dAction = {}
                        dAction['constant'] = sTargetValue
                        dAction['found'] = sTokenValue
                        oViolation.set_action(dAction)
                        self.add_violation(oViolation)

    def _fix_violation(self, oViolation):
        lTokens = oViolation.get_tokens()
        dActions = oViolation.get_action()
        lTokens[0].set_value(dActions['constant'])
        oViolation.set_tokens(lTokens)


def is_token_in_ignore_token_list(oToken, lIgnoreTokens):
    for oIgnoreToken in lIgnoreTokens:
        if isinstance(oToken, oIgnoreToken):
            return True
    return False




from vsg.rules import token_indent

from vsg import token

lTokens = []
lTokens.append(token.enumeration_type_definition.close_parenthesis)


class rule_016(token_indent):
    '''
    This rule checks the indent of the closing parenthesis on multiline types.

    **Violation**

    .. code-block:: vhdl

       architecture rtl of fifo is

         type state_machine is (
           idle, write, read, done
           );

       begin

    **Fix**

    .. code-block:: vhdl

       architecture rtl of fifo is

         type state_machine is (
           idle, write, read, done
         );

       begin
    '''

    def __init__(self):
        token_indent.__init__(self, 'type', '016', lTokens)




from vsg import parser
from vsg import token

from vsg.rules import consistent_token_case

lTokens = []
lTokens.append(token.function_specification.designator)

lIgnore = []
lIgnore.append(parser.whitespace)
lIgnore.append(parser.carriage_return)
lIgnore.append(parser.blank_line)


class rule_010(consistent_token_case):
    '''
    This rule checks for consistent capitalization of function names.

    **Violation**

    .. code-block:: vhdl

       architecture rtl of fifo is

         function func_1 ()

       begin

         OUT1 <= Func_1;

         PROC1 : process () is
         begin

            sig1 <= FUNC_1;

         end process;

       end architecture rtl;

    **Fix**

    .. code-block:: vhdl

       architecture rtl of fifo is

         function func_1 ()

       begin

         OUT1 <= func_1;

         PROC1 : process () is
         begin

            sig1 <= func_1;

         end process;

       end architecture rtl;
    '''

    def __init__(self):
        consistent_token_case.__init__(self, 'function', '010', lTokens, lIgnore)
